"""Common docstrings."""

from __future__ import annotations

from module_utilities.docfiller import DocFiller


def _docstrings_shared() -> None:  # pylint: disable=differing-param-doc,differing-type-doc
    """
    Parameters
    ----------
    segments : sequence of int
        Integration limits. For ``n = len(segments)`` integration will be performed over ranges
        ``(segments[0], segments[1]), (segments[1], segments[2]), ..., (segments[n-2], segments[n-]])``
    phi_rep : callable
        Repulsive part of pair potential.
    beta : float
        Inverse temperature.
    phi : callable
        Potential function.
    r : float or array-like
        Pair separation distance(s).
    quad_kws : mapping, optional
        Extra arguments to :func:`analphipy.utils.quad_segments`
    r_min_exact | r_min : float
        Location of minimum in potential energy.
    phi_min_exact | phi_min : float, optional
        Value of potential energy at minimum.
    full_output : bool, optional
        If True, return extra information.
    err : bool, optional
        If True, return error.
    error_summed | error : float, optional
        Total integration error. Returned if ``err`` or ``full_output`` are `True`.
    full_output_summed | outputs : object
        Output(s) from :func:`scipy.integrate.quad`.  Returned if ``full_output`` is True.
    volume_int_func | volume : str or callable, optional
        Volume element in integration.
        For example, use ``volume = lambda x: 4 * np.pi * x ** 2`` for spherically symmetric 3d integration.
        Can also pass string value of {'1d', '2d', '3d'}.  If passed None, then assume '1d' integration.

    """


_references = {
    "ref_Noro_Frenkel": "M.G. Noro and D. Frenkel (2000), 'Extended corresponding-states behavior for particles with variable range attractions'. Journal of Chemical Physics, 113, 2941.",
    "ref_Barker_Henderson": "J.A. Barker and D. Henderson (1976), 'What Is Liquid? Understanding the States of Matter'. Reviews of Modern Physics, 48, 587-671",
    "ref_WCA": "J.D. Weeks, D. Chandler and H.C. Andersen (1971), 'Role of Repulsive Forces in Determining the Equilibrium Structure of Simple Liquids', Journal of Chemical Physics 54, 5237-5247",
    "kl_link": "`See here for more info on Kullback & Leibeler divergence <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence#Definition>`_",
    "js_link": "`See here for more info on Jeffreys symmetric divergence <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence#Symmetrised_divergence>`_",
}


docfiller = DocFiller.from_docstring(
    _docstrings_shared, combine_keys="parameters"
).update(**_references)



from __future__ import annotations

from typing import TYPE_CHECKING

from analphipy import potential as pots
from analphipy.norofrenkel import NoroFrenkelPair

if TYPE_CHECKING:
    from collections.abc import Iterator

    import pandas as pd

    from analphipy.base_potential import PhiAbstract


def iter_phi_lj(
    table: pd.DataFrame,
    nsamp: int,
) -> Iterator[tuple[pd.DataFrame, NoroFrenkelPair]]:
    if len(table) > nsamp:
        table = table.sample(nsamp)
    sig = eps = 1.0
    p_base = pots.LennardJones(sig=sig, eps=eps)

    p: PhiAbstract

    for (rcut, tail), g in table.groupby(["rcut", "tail"]):
        if tail == "LFS":
            p = p_base.lfs(rcut=rcut)
        elif tail == "CUT":
            p = p_base.cut(rcut=rcut)
        elif tail == "LRC":
            p = p_base

        a = NoroFrenkelPair.from_phi(
            phi=p.phi, segments=p.segments, r_min=sig, bounds=[0.5, 1.5]
        )

        yield g, a



"""
Utilities module (:mod:`analphipy.utils`)
=========================================
"""

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, cast

import numpy as np

from ._docstrings import docfiller

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping
    from typing import Any, Protocol, TypeVar

    from ._typing import ArrayLike, OptimizeResultInterface, P, QuadSegments, R
    from ._typing_compat import Concatenate, TypeGuard


TWO_PI: float = 2.0 * np.pi


def combine_segmets(a: ArrayLike, b: ArrayLike) -> list[float]:
    """
    Combine two lists of segments.

    Parameters
    ----------
    a, b : array-like
        list of segments.

    Returns
    -------
    combined : list of float
        sorted unique values for ``a`` and ``b``.

    """
    aa, bb = set(a), set(b)
    return sorted(aa.union(bb))  # type: ignore[arg-type, unused-ignore]


def is_float(val: Any) -> TypeGuard[float]:
    """
    Type guard for float.

    Use to type narrow output for quad_segments
    """
    return isinstance(val, float)


@docfiller.decorate
def quad_segments(
    func: Callable[..., Any],
    segments: ArrayLike,
    args: tuple[Any, ...] = (),
    full_output: bool = False,
    sum_integrals: bool = True,
    sum_errors: bool = False,
    err: bool = True,
    **kws: Any,
) -> QuadSegments:
    """
    Perform quadrature with discontinuities.

    Parameters
    ----------
    func : callable
        function to be integrated
    {segments}
    args : tuple, optional
        Extra positional arguments to `func`.
    full_output : bool, default=False
        If True, return extra information.
    sum_integrals : bool, default=True
        If True, sum the segments in the output.
    sum_errors : bool, default=True
        If True and returning `error` sum errors.
    err : bool, default=True
        If True, return error.
    **kws :
        Extra arguments to :func:`scipy.integrate.quad`

    Returns
    -------
    integral : float or list of float
        If `sum_integrals`,  this is the sum of integrals over each segment.  Otherwise return list
        of values corresponding to integral in each segment.
    errors : float or list of float, optional
        If `err` or `full_output` are True, then return error.  If `sum_errors`, then sum of errors
        Across segments.
    outputs : object
        Output from :func:`scipy.integrate.quad`. If multiple segments, return a list of output.

    See Also
    --------
    scipy.integrate.quad

    """
    from scipy.integrate import (  # pyright: ignore[reportMissingTypeStubs]
        quad,  # pyright: ignore[reportUnknownVariableType]
    )

    out: list[tuple[float, float, dict[str, Any]]] = [
        quad(func, a=a, b=b, args=args, full_output=True, **kws)
        for a, b in zip(segments[:-1], segments[1:])
    ]

    integrals: float | list[float]
    errors: float | list[float]
    outputs: dict[str, Any] | list[dict[str, Any]]

    if len(segments) == 2:  # noqa: PLR2004
        integrals, errors, outputs = out[0]
    else:
        integrals_list: list[float] = []
        errors_list: list[float] = []
        outputs_list: list[dict[str, Any]] = []

        for y, e, o in out:
            integrals_list.append(y)
            errors_list.append(e)
            outputs_list.append(o)

        integrals = (
            cast("float", np.sum(integrals_list))  # pyright: ignore[reportUnknownMemberType]
            if sum_integrals
            else integrals_list
        )

        errors = (
            np.sum(errors_list)  # pyright: ignore[reportUnknownMemberType]
            if sum_errors
            else errors_list
        )

        outputs = outputs_list

    # Gather final results
    if err and full_output:
        return integrals, errors, outputs
    if err and not full_output:
        return integrals, errors
    if not err and full_output:
        return integrals, outputs
    return integrals


def minimize_phi(
    phi: Callable[..., Any], r0: float, bounds: ArrayLike | None = None, **kws: Any
) -> tuple[float, float, OptimizeResultInterface | None]:
    """
    Find value of ``r`` which minimized ``phi``.

    Parameters
    ----------
    phi : callable
        Function to be minimized.
    r0 : float
        Guess for position of minimum.
    bounds : tuple of float
        If passed, should be of form ``bounds=(lower_bound, upper_bound)``.
    **kws :
        Extra arguments to :func:`scipy.optimize.minimize`

    Returns
    -------
    rmin : float
        ``phi(rmin)`` is found location of minimum.
    phimin : float
        Value of ``phi(rmin)``, i.e., the value of ``phi`` at the minimum
    output : object
        Output class from :func:`scipy.optimize.minimize`.

    See Also
    --------
    scipy.optimize.minimize

    """
    from scipy.optimize import (  # pyright: ignore[reportMissingTypeStubs]
        minimize,  # pyright: ignore[reportUnknownVariableType]
    )

    if bounds is None:
        bounds = (0.0, np.inf)

    if bounds[0] == bounds[1]:
        # force minima to be exactly here
        xmin = bounds[0]
        ymin = phi(xmin)
        return xmin, ymin, None

    outputs = cast("OptimizeResultInterface", minimize(phi, r0, bounds=[bounds], **kws))

    if not outputs["success"]:
        msg = "could not find min of phi"
        raise ValueError(msg)

    xmin = outputs["x"][0]

    tmp = outputs["fun"]
    ymin = tmp[0] if isinstance(tmp, np.ndarray) else tmp

    return cast("tuple[float, float, OptimizeResultInterface]", (xmin, ymin, outputs))


# * Phi utilities
if TYPE_CHECKING:

    class HasQuadKws(Protocol):
        """Class protocol for quad_kws"""

        quad_kws: Mapping[str, Any]

    S = TypeVar("S", bound=HasQuadKws)


def partial_phi(phi: Callable[..., R], **params: Any) -> Callable[..., R]:
    """Partial potential with params set."""
    from functools import partial

    return partial(phi, **params)


def segments_to_segments_cut(segments: Iterable[float], rcut: float) -> list[float]:
    """Update segments for 'cut' potential."""
    return [x for x in segments if x < rcut] + [rcut]


def add_quad_kws(
    func: Callable[Concatenate[S, P], R],
) -> Callable[Concatenate[S, P], R]:
    """Add quad keyword arguments."""

    @wraps(func)
    def wrapped(self: S, /, *args: P.args, **kws: P.kwargs) -> R:
        kws = dict(self.quad_kws, **kws)  # type: ignore[assignment]
        return func(self, *args, **kws)

    return wrapped



#
# python_boilerplate documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  9 13:47:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#
"""Build docs."""

from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import Any

sys.path.insert(0, str(Path("../src").resolve()))

import analphipy

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "autodocsumm",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    # "IPython.sphinxext.ipython_directive",
    # "IPython.sphinxext.ipython_console_highlighting",
    # "nbsphinx",
    # - easier external links
    # "sphinx.ext.extlinks",
    # - view source code on created page
    # "sphinx.ext.viewcode",
    # - view source code on github
    "sphinx.ext.linkcode",
    # - add copy button
    "sphinx_copybutton",
    # - redirect stuff?
    # "sphinxext.rediraffe",
    # - pretty things up?
    # "sphinx_design"
    # - myst stuff
    "myst_nb",
    # "myst_parser",
]

nitpicky = True
autosectionlabel_prefix_document = True

# -- myst stuff ---------------------------------------------------------
myst_enable_extensions = [
    "dollarmath",
    "amsmath",
    "deflist",
    "fieldlist",
    "html_admonition",
    "html_image",
    "colon_fence",
    "smartquotes",
    "replacements",
    # "linkify",
    "strikethrough",
    "substitution",
    "tasklist",
    # "attrs_inline",
    # "attrs_block",
]


myst_heading_anchors = 3
myst_footnote_transition = True
myst_dmath_double_inline = True
myst_enable_checkboxes = True
myst_substitutions = {
    "role": "[role](#syntax/roles)",
    "directive": "[directive](#syntax/directives)",
}
# myst_enable_extensions = [
#     "dollarmath",
#     "amsmath",
#     "deflist",
#     # "html_admonition",
#     "html_image",
#     "colon_fence",
#     # "smartquotes",
#     # "replacements",
#     # "linkify",
#     # "substitution",
#     "attrs_inline",
#     "attrs_block",
# ]

myst_url_schemes = ("http", "https", "mailto")

nb_execution_mode = "cache"
# nb_execution_mode = "auto"

# set the kernel name
nb_kernel_rgx_aliases = {
    "analphipy.*": "python3",
    "conda.*": "python3",
}

nb_execution_allow_errors = True

# Whether to remove stderr
nb_output_stderr = "remove"

# - top level variables --------------------------------------------------------
# set github_username variable to be subbed later.
# this makes it easy to switch from wpk -> usnistgov later
github_username = "usnistgov"

html_context = {
    "github_user": github_username,
    "github_repo": "analphipy",
    "github_version": "main",
    "doc_path": "docs",
}

# -- python3 ---------------------------------------------------------------
autosummary_generate = True
# autosummary_generate = False
autodoc_member_order = "bysource"

# autoclass_content = "both"  # include both class docstring and __init__
autodoc_default_flags = [
    # Make sure that any autodoc declarations show the right members
    "members",
    "inherited-members",
    "private-members",
    "show-inheritance",
]
autodoc_typehints = "none"

# -- napoleon ------------------------------------------------------------------
napoleon_google_docstring = False
napoleon_numpy_docstring = True

napoleon_use_param = False
napoleon_use_rtype = False
napoleon_preprocess_types = True
napoleon_type_aliases = {
    # general terms
    "sequence": ":term:`sequence`",
    "iterable": ":term:`iterable`",
    "callable": ":py:func:`callable`",
    "dict_like": ":term:`dict-like <mapping>`",
    "dict-like": ":term:`dict-like <mapping>`",
    "path-like": ":term:`path-like <path-like object>`",
    "mapping": ":term:`mapping`",
    "hashable": ":term:`hashable`",
    "file-like": ":term:`file-like <file-like object>`",
    # special terms
    # "same type as caller": "*same type as caller*",  # does not work, yet
    # "same type as values": "*same type as values*",  # does not work, yet
    # stdlib type aliases
    "MutableMapping": "~collections.abc.MutableMapping",
    "sys.stdout": ":obj:`sys.stdout`",
    "timedelta": "~datetime.timedelta",
    "string": ":class:`string <str>`",
    # numpy terms
    "array_like": ":term:`array_like`",
    "array-like": ":term:`array-like <array_like>`",
    "scalar": ":term:`scalar`",
    "array": ":term:`array`",
    # matplotlib terms
    "color-like": ":py:func:`color-like <matplotlib.colors.is_color_like>`",
    "matplotlib colormap name": ":doc:`matplotlib colormap name <matplotlib:gallery/color/colormap_reference>`",
    "matplotlib axes object": ":py:class:`matplotlib axes object <matplotlib.axes.Axes>`",
    "colormap": ":py:class:`colormap <matplotlib.colors.Colormap>`",
    # objects without namespace: xarray
    "DataArray": "~xarray.DataArray",
    "Dataset": "~xarray.Dataset",
    "Variable": "~xarray.Variable",
    "DatasetGroupBy": "~xarray.core.groupby.DatasetGroupBy",
    "DataArrayGroupBy": "~xarray.core.groupby.DataArrayGroupBy",
    "CentralMoments": "~cmomy.CentralMoments",
    "xCentralMoments": "~cmomy.xCentralMoments",
    # objects without namespace: numpy
    "ndarray": "~numpy.ndarray",
    "MaskedArray": "~numpy.ma.MaskedArray",
    "dtype": "~numpy.dtype",
    "ComplexWarning": "~numpy.ComplexWarning",
    # objects without namespace: pandas
    "Index": "~pandas.Index",
    "MultiIndex": "~pandas.MultiIndex",
    "CategoricalIndex": "~pandas.CategoricalIndex",
    "TimedeltaIndex": "~pandas.TimedeltaIndex",
    "DatetimeIndex": "~pandas.DatetimeIndex",
    "Series": "~pandas.Series",
    "DataFrame": "~pandas.DataFrame",
    "Categorical": "~pandas.Categorical",
    "Path": "~~pathlib.Path",
    # objects with abbreviated namespace (from pandas)
    "pd.Index": "~pandas.Index",
    "pd.NaT": "~pandas.NaT",
}


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
    ".myst": "myst-nb",
}

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "analphipy"
copyright = "2021, William P. Krekelberg"  # noqa: A001
author = "William P. Krekelberg"


# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
def _get_version() -> str:
    if (version := os.environ.get("SETUPTOOLS_SCM_PRETEND_VERSION")) is None:
        version = analphipy.__version__
    return version


release = version = _get_version()


# if always want to print "latest"
# release = "latest"
# version = "latest"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = "sphinx_book_theme"

html_theme_options = {
    # "analytics_id": ''  this is configured in rtfd.io
    # "canonical_url": "",
    "repository_url": f"https://github.com/{github_username}/analphipy",
    "repository_branch": html_context["github_version"],
    "path_to_docs": html_context["doc_path"],
    # "use_edit_page_button": True,
    "use_repository_button": True,
    "use_issues_button": True,
    "home_page_in_toc": True,
    "show_toc_level": 2,
    "show_navbar_depth": 0,
}
# handle nist css/js from here.
html_css_files = [
    # "css/nist-combined.css",
    "https://pages.nist.gov/nist-header-footer/css/nist-combined.css",
    "https://pages.nist.gov/leaveNotice/css/jquery.leaveNotice.css",
]

html_js_files = [
    "https://code.jquery.com/jquery-3.6.2.min.js",
    "https://pages.nist.gov/nist-header-footer/js/nist-header-footer.js",
    # "js/nist-header-footer.js",
    "https://pages.nist.gov/leaveNotice/js/jquery.leaveNotice-nist.min.js",
    "js/leave_notice.js",
    # google stuff:
    (
        "https://dap.digitalgov.gov/Universal-Federated-Analytics-Min.js?agency=NIST&subagency=github&pua=UA-66610693-1&yt=true&exts=ppsx,pps,f90,sch,rtf,wrl,txz,m1v,xlsm,msi,xsd,f,tif,eps,mpg,xml,pl,xlt,c",
        {"async": "async", "id": "_fed_au_ua_tag", "type": "text/javascript"},
    ),
]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# Sometimes the savefig directory doesn't exist and needs to be created
# https://github.com/ipython/ipython/issues/8733
# becomes obsolete when we can pin ipython>=5.2; see ci/requirements/doc.yml
def _get_ipython_savefig_dir() -> str:
    d = Path(__file__).parent / "_build" / "html" / "_static"
    if not d.is_dir():
        d.mkdir(parents=True)
    return str(d)


ipython_savefig_dir = _get_ipython_savefig_dir()

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
today_fmt = "%Y-%m-%d"
html_last_updated_fmt = today_fmt


# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "analphipydoc"


# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "analphipy.tex",
        "analphipy Documentation",
        "William P. Krekelberg",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "analphipy",
        "analphipy Documentation",
        [author],
        1,
    ),
]


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "analphipy",
        "analphipy Documentation",
        author,
        "analphipy",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# -- user defined stuff ------------------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "numba": ("https://numba.pydata.org/numba-doc/latest", None),
    # "matplotlib": ("https://matplotlib.org", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "dask": ("https://docs.dask.org/en/latest", None),
    "cftime": ("https://unidata.github.io/cftime", None),
    "sparse": ("https://sparse.pydata.org/en/latest/", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
}

linkcheck_ignore = ["https://doi.org/"]


# based on numpy doc/source/conf.py
def linkcode_resolve(domain: str, info: dict[str, Any]) -> str | None:
    """Determine the URL corresponding to Python object."""
    import inspect
    from operator import attrgetter

    if domain != "py":
        return None

    parent_name, *sub_parts = info["module"].split(".")
    parent_mod = sys.modules.get(parent_name)

    try:
        if len(sub_parts) > 0:
            sub_name = ".".join(sub_parts)
            obj = attrgetter(sub_name)(parent_mod)
        else:
            obj = parent_mod

        # get fullname
        obj = attrgetter(info["fullname"])(obj)

    except AttributeError:
        return None

    try:
        fn = inspect.getsourcefile(inspect.unwrap(obj))
    except TypeError:
        fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except OSError:
        lineno = None

    linespec = f"#L{lineno}-L{lineno + len(source) - 1}" if lineno else ""

    # fmt: off
    fn = os.path.relpath(fn, start=Path(analphipy.__file__).parent)
    # fmt: on

    return f"https://github.com/{github_username}/analphipy/blob/{html_context['github_version']}/src/analphipy/{fn}{linespec}"


# only set spelling stuff if installed:
try:
    import sphinxcontrib.spelling  # noqa: F401

    extensions += ["sphinxcontrib.spelling"]
    spelling_word_list_filename = "spelling_wordlist.txt"

except ImportError:
    pass



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
# pylint: disable=missing-class-docstring,duplicate-code
from __future__ import annotations

from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Any

import numpy as np
import pytest
from typing_extensions import Self

if TYPE_CHECKING:
    from collections.abc import Iterable

rng = np.random.default_rng()


def phidphi_lj(r, sig=1.0, eps=1.0):
    """
    Lennard jones potential and radial force.

    Returns
    -------
    phi : array
        potential at `r`
    dphi : array
        value of -d(phi)/d(r) * 1/r at `r`
    """

    r = np.array(r)

    rinvsq = 1.0 / (r * r)

    x2 = (sig * sig) * rinvsq
    x6 = x2 * x2 * x2

    phi = 4.0 * eps * x6 * (x6 - 1.0)
    dphi = 48.0 * eps * x6 * (x6 - 0.5) * rinvsq

    return phi, dphi


def phidphi_lj_cut(r, rcut, sig=1.0, eps=1.0):
    r = np.array(r)
    phi = np.empty_like(r)
    dphi = np.empty_like(r)

    m = r <= rcut
    phi[~m] = 0.0
    dphi[~m] = 0.0

    if np.any(m):
        v, dv = phidphi_lj(r[m], sig=sig, eps=eps)
        vcut, _dvcut = phidphi_lj(rcut, sig=sig, eps=eps)

        phi[m] = v - vcut
        dphi[m] = dv

    return phi, dphi


def phidphi_lj_lfs(r, rcut, sig=1.0, eps=1.0):
    r = np.array(r)
    phi = np.empty_like(r)
    dphi = np.empty_like(r)

    m = r <= rcut
    phi[~m] = 0.0
    dphi[~m] = 0.0

    if np.any(m):
        v, dv = phidphi_lj(r[m], sig=sig, eps=eps)
        vcut, dvcut = phidphi_lj(rcut, sig=sig, eps=eps)

        phi[m] = v - vcut + dvcut * rcut * (r[m] - rcut)
        dphi[m] = dv - dvcut * rcut / r[m]
    return phi, dphi


def _prefac_nm(n, m, eps):
    return eps * (n / (n - m)) * (n / m) ** (m / (n - m))


def phidphi_nm(r, n, m, sig=1.0, eps=1.0):
    r = np.array(r)
    prefac = _prefac_nm(n=n, m=m, eps=eps)

    x = sig / r

    xn = x**n
    xm = x**m
    phi = prefac * (xn - xm)

    # dphi = -1/r dphi/dr = x dphi/dx * 1/r**2
    # where x = sig / r
    dphi = prefac * (n * xn - m * xm) / r**2

    return phi, dphi


def get_r(rmin, rmax, n=100):
    rmin += rng.random() * 0.1
    rmax += rng.random() * 0.1

    return np.linspace(rmin, rmax, n)


def kws_to_ld(**kws: Iterable[Any]) -> list[dict[str, Any]]:
    return [dict(zip(kws, t)) for t in zip(*kws.values())]


@dataclass
class BaseParams:
    def __post_init__(self):
        self.r = None

    def phidphi(self, r) -> None:
        raise NotImplementedError

    def get_phidphi(self):
        if hasattr(self, "phidphi"):
            return self.phidphi(self.r)
        raise NotImplementedError

    def phi(self, r):
        return self.phidphi(r)[0]  # type: ignore[func-returns-value]

    def get_phi(self):
        return self.phi(self.r)

    @classmethod
    def get_params(cls, nsamp) -> None:
        raise NotImplementedError

    @classmethod
    def get_objects(cls, nsamp=10) -> Iterable[Self]:
        for params in cls.get_params(nsamp=nsamp):  # type: ignore[attr-defined]
            yield cls(**params)

    def asdict(self):
        return asdict(self)


@dataclass
class LJParams(BaseParams):
    sig: float
    eps: float

    def __post_init__(self):
        self.r = get_r(rmin=0.1 * self.sig, rmax=5.0 * self.sig, n=100)

    def phidphi(self, r):
        return phidphi_lj(r, sig=self.sig, eps=self.eps)

    @classmethod
    def get_params(cls, nsamp=10):
        return kws_to_ld(
            sig=rng.random(nsamp) * 4,
            eps=rng.random(nsamp),
        )


@dataclass
class LJCutParams(LJParams):
    rcut: float

    def __post_init__(self):
        self.r = get_r(rmin=0.1 * self.sig, rmax=self.rcut + self.sig, n=100)

    def phidphi(self, r):
        return phidphi_lj_cut(r, sig=self.sig, eps=self.eps, rcut=self.rcut)

    @classmethod
    def get_params(cls, nsamp=10):
        return kws_to_ld(
            sig=rng.random(nsamp),
            eps=rng.random(nsamp),
            rcut=rng.random(nsamp) + 3.0,
        )


@dataclass
class LJLFSParams(LJCutParams):
    def phidphi(self, r):
        return phidphi_lj_lfs(r, sig=self.sig, eps=self.eps, rcut=self.rcut)


@pytest.fixture(params=LJParams.get_objects(), scope="module")
def lj_params(request):
    return request.param


@pytest.fixture(params=LJCutParams.get_objects(), scope="module")
def lj_cut_params(request):
    return request.param


@pytest.fixture(params=LJLFSParams.get_objects(), scope="module")
def lj_lfs_params(request):
    return request.param


@dataclass
class NMParams(LJParams):
    n: int
    m: int

    def phidphi(self, r):
        return phidphi_nm(r, n=self.n, m=self.m, sig=self.sig, eps=self.eps)

    @classmethod
    def get_params(cls, nsamp=None):
        n = [8, 10, 12, 18]
        m = [6, 7, 6, 9]

        if nsamp is None:
            nsamp = len(n)

        return kws_to_ld(sig=rng.random(nsamp), eps=rng.random(nsamp), n=n, m=m)


@pytest.fixture(params=NMParams.get_objects(), scope="module")
def nm_params(request):
    return request.param


def phi_sw(r, sig=1.0, eps=-1.0, lam=1.5):
    """
    Square well potential with value.

    * inf : r < sig
    * eps : sig <= r < lam * sig
    * 0 :  r >= lam * sig
    """

    r = np.array(r)

    phi = np.empty_like(r)

    m0 = r < sig
    m1 = (r >= sig) & (r < lam * sig)
    m2 = r >= lam * sig

    phi[m0] = np.inf
    phi[m1] = eps
    phi[m2] = 0.0

    return phi


@dataclass
class SWParams(LJParams):
    lam: float

    def phi(self, r):
        return phi_sw(r, sig=self.sig, eps=self.eps, lam=self.lam)

    @classmethod
    def get_params(cls, nsamp=10):
        return kws_to_ld(
            sig=rng.random(nsamp),
            eps=(rng.random(nsamp) - 0.5) * 2.0,
            lam=rng.random(nsamp) + 1.0,
        )


@pytest.fixture(params=SWParams.get_objects(), scope="module")
def sw_params(request):
    return request.param


def phi_yk(r, z, sig=1.0, eps=1.0):
    r = np.array(r)
    phi = np.empty_like(r)
    m = r >= sig

    phi[~m] = np.inf
    if np.any(m):
        x = r[m] / sig
        phi[m] = -eps * np.exp(-z * (x - 1.0)) / x
    return phi


@dataclass
class YKParams(LJParams):
    z: float

    def phi(self, r):
        return phi_yk(r, z=self.z, sig=self.sig, eps=self.eps)

    @classmethod
    def get_params(cls, nsamp=10):
        return kws_to_ld(
            sig=rng.random(nsamp),
            eps=(rng.random(nsamp) - 0.5) * 2.0,
            z=rng.random(nsamp) * 4,
        )


@pytest.fixture(params=YKParams.get_objects(), scope="module")
def yk_params(request):
    return request.param


def phi_hs(r, sig=1.0):
    r = np.array(r)

    phi = np.empty_like(r)

    m0 = r < sig
    phi[m0] = np.inf
    phi[~m0] = 0.0

    return phi


@dataclass
class HSParams(BaseParams):  # pylint: disable=abstract-method
    sig: float

    def __post_init__(self):
        self.r = get_r(rmin=0.1 * self.sig, rmax=2 * self.sig, n=100)

    def phi(self, r):  # noqa: ARG002
        return phi_hs(self.r, sig=self.sig)

    @classmethod
    def get_params(cls, nsamp=10):
        return kws_to_ld(sig=rng.random(nsamp))


@pytest.fixture(params=HSParams.get_objects(), scope="module")
def hs_params(request):
    return request.param



# pylint: disable=redefined-builtin
"""
Light weight argparser from a dataclass.

There are some libraries out there that mostly do what we want, but none are quite right.


Took motivation from the following:
https://github.com/typed-argparse/typed-argparse
https://github.com/SunDoge/typed-args
https://github.com/rmorshea/noxopt


To get a parser use something like:

from typing import TypeAlias
RUN_TYPE: TypeAlias = list[list[str]] | None


@dataclass
class Example(DataclassParser):
    cmd: list[Literal['a','b']] | None = add_option("-c", "--cmd")
    run: Annotated[list[list[str]] | None, Option(help='hello')] = add_option("-r", "--run")
    another: Annotated[list[list[str]] | None, option("-a", "--another", help='hello')] = None

    other: RUN_TYPE = add_option(help="hello")
    version: str | None = None
    lock: bool = False
    no_cov: bool = False

"""

from __future__ import annotations

import sys
from argparse import ArgumentParser
from dataclasses import (
    dataclass,
    field,
    fields,
    is_dataclass,
    replace,
)
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Literal,
    Union,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Container, Sequence

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


if sys.version_info < (3, 10):
    msg = "Require python >= 3.10"
    raise RuntimeError(msg)


_NoneType = type(None)

UNDEFINED = cast(
    "Any",
    type("Undefined", (), {"__repr__": lambda self: "UNDEFINED"})(),  # pyright: ignore[reportUnknownLambdaType]  # noqa: ARG005
)


@dataclass
class Option:
    """Class to handle options."""

    flags: Sequence[str] = UNDEFINED
    # remainder are alphabetical
    action: str | None = UNDEFINED
    choices: Container[Any] = UNDEFINED
    const: Any = UNDEFINED
    default: Any | None = UNDEFINED
    dest: str = UNDEFINED
    help: str = UNDEFINED
    metavar: str = UNDEFINED
    nargs: str | int | None = UNDEFINED
    required: bool = UNDEFINED
    type: int | float | Callable[[Any], Any] = UNDEFINED
    prefix_char: str = "-"

    def __post_init__(self) -> None:
        if isinstance(self.flags, str):
            self.flags = (self.flags,)
        if self.flags is not UNDEFINED:
            for f in self.flags:
                if not f.startswith(self.prefix_char):
                    msg = f"Option only supports flags, but got {f!r}"
                    raise ValueError(msg)

    def asdict(self) -> dict[str, Any]:
        """Convert to dictionary."""
        return {
            k: v
            for k, v in (
                # Can't use asdict() since that deep copies and we need
                # to filter using an identity check against UNDEFINED.
                [
                    (f.name, getattr(self, f.name))
                    for f in fields(self)
                    if f.name != "prefix_char"
                ]
            )
            if v is not UNDEFINED
        }

    def add_argument_to_parser(
        self,
        parser: ArgumentParser,
        prefix_char: str = "-",
    ) -> None:
        """Add argument to parser."""
        kwargs = self.asdict()

        flags = kwargs.pop("flags")

        # make sure flags have correct prefixing:
        if not all(flag.startswith(prefix_char) for flag in flags):
            new_flags: list[str] = []
            for flag in flags:
                if flag.startswith(prefix_char):
                    new_flags.append(flag)
                elif flag.startswith("--"):
                    new_flags.append(prefix_char * 2 + flag.lstrip("-"))
                elif flag.startswith("-"):
                    new_flags.append(prefix_char + flag.lstrip("-"))
                else:
                    msg = f"bad flag {flag} prefix_char {prefix_char}"
                    raise ValueError(msg)

            flags = new_flags

        parser.add_argument(*flags, **kwargs)

    @classmethod
    def factory(
        cls,
        *flags: str,
        action: str | None = UNDEFINED,
        choices: Container[Any] = UNDEFINED,
        const: Any = UNDEFINED,
        default: Any | None = UNDEFINED,
        dest: str = UNDEFINED,
        help: str = UNDEFINED,
        metavar: str = UNDEFINED,
        nargs: str | int | None = UNDEFINED,
        required: bool = UNDEFINED,
        type: Callable[[Any], Any] = UNDEFINED,
    ) -> Self:
        """Factory method."""
        return cls(
            flags=flags or UNDEFINED,
            action=action,
            choices=choices,
            const=const,
            default=default,
            dest=dest,
            help=help,
            metavar=metavar,
            nargs=nargs,
            required=required,
            type=type,
        )


option = Option.factory


def add_option(
    *flags: str,
    default: Any = None,
    action: str | None = UNDEFINED,
    choices: Container[Any] = UNDEFINED,
    const: Any = UNDEFINED,
    dest: str = UNDEFINED,
    help: str = UNDEFINED,
    metavar: str = UNDEFINED,
    nargs: str | int | None = UNDEFINED,
    required: bool = UNDEFINED,
    type: Callable[[Any], Any] = UNDEFINED,
    **field_kws: Any,  # noqa: ARG001
) -> Any:
    """Add option."""
    return field(  # pylint: disable=invalid-field-call
        metadata={
            "option": Option.factory(
                *flags,
                default=default,
                action=action,
                choices=choices,
                const=const,
                dest=dest,
                help=help,
                metavar=metavar,
                nargs=nargs,
                required=required,
                type=type,
            ),
        },
        default=default,
    )


class DataclassParser:
    """Mixin for creating a dataclass with parsing."""

    @classmethod
    def parser(cls, prefix_char: str = "-", **kwargs: Any) -> ArgumentParser:
        """Get parser."""
        parser = ArgumentParser(prefix_chars=prefix_char, **kwargs)

        for opt in get_dataclass_options(cls).values():
            opt.add_argument_to_parser(parser, prefix_char=prefix_char)

        return parser

    @classmethod
    def from_posargs(
        cls,
        posargs: str | Sequence[str],
        prefix_char: str = "-",
        parser: ArgumentParser | None = None,
        known: bool = False,
    ) -> Self:
        """Create object from posargs."""
        if parser is None:
            parser = cls.parser(prefix_char=prefix_char)

        if isinstance(posargs, str):
            import shlex

            posargs = shlex.split(posargs)

        if known:
            parsed, _ = parser.parse_known_args(posargs)
        else:
            parsed = parser.parse_args(posargs)

        return cls(**vars(parsed))


def get_dataclass_options(cls: Any) -> dict[str, Option]:
    """Get dictionary of options."""
    return {
        name: _create_option(name=name, opt=opt, annotation=annotation)
        for name, (annotation, opt) in _get_dataclass_annotations_and_options(
            cls,
        ).items()
    }


def _get_dataclass_annotations_and_options(
    cls: Any,
) -> dict[str, tuple[Any, Option]]:
    annotations = get_type_hints(cls, include_extras=True)

    assert is_dataclass(cls)
    cls_fields = fields(cls)

    out: dict[str, tuple[Any, Any]] = {}
    for f in cls_fields:
        if f.name.startswith("_") or not f.init:
            continue

        opt = f.metadata.get("option", Option(default=f.default))

        # Can also pass via annotations
        annotation = annotations[f.name]
        if get_origin(annotation) is Annotated:
            annotation, opt_anno, *_ = get_args(annotation)
            if isinstance(opt_anno, Option):
                opt = Option(**{**opt_anno.asdict(), **opt.asdict()})

        out[f.name] = (annotation, opt)
    return out


def _create_option(
    name: str,
    opt: Option,
    annotation: Any,
) -> Option:
    depth, underlying_type = _get_underlying_type(annotation)
    max_depth = 2

    if depth <= max_depth and get_origin(underlying_type) is Literal:
        choices = get_args(underlying_type)
        if opt.choices is UNDEFINED:
            opt = replace(opt, choices=choices)
    else:
        choices = None

    if opt.nargs is UNDEFINED and depth > 0:
        opt = replace(opt, nargs="*")

    if opt.action is UNDEFINED and depth > 1:
        opt = replace(opt, action="append")

    if opt.type is UNDEFINED:
        opt_type = type(choices[0]) if choices else underlying_type  # pyright: ignore[reportUnknownVariableType]

        if not callable(opt_type):  # pyright: ignore[reportUnknownArgumentType]
            msg = (
                f"Annotation {annotation} for parameter {name!r} is not callable."
                f"Declare arg type with Annotated[..., Option(type=...)] instead."
            )
            raise TypeError(
                msg,
            )
        opt = replace(opt, type=opt_type)

    if opt.type is bool:
        opt = replace(
            opt,
            action="store_false" if opt.default is True else "store_true",
            type=UNDEFINED,
            default=UNDEFINED,
        )

    if opt.flags is UNDEFINED or not opt.flags:
        opt = replace(opt, flags="--" + name.replace("_", "-"))

    return opt


def _get_underlying_type(
    opt: Any,
    allow_optional: bool = True,
    depth: int = 0,
    max_depth: int = 2,
) -> tuple[int, Any]:
    """
    Parse nested list of type -> (depth, type)

    list[list[str]] -> (2, str)
    """
    depth_out = depth

    type_: Any = opt

    if depth > max_depth or ((opt_origin := get_origin(opt)) is None):
        pass

    elif opt_origin is list:
        opt_arg, *extras = get_args(opt)
        if not extras:
            depth_out, type_ = _get_underlying_type(
                opt_arg,
                allow_optional=False,
                depth=depth + 1,
                max_depth=max_depth,
            )

    elif allow_optional and (underlying := _get_underlying_if_optional(opt)):  # pylint: disable=confusing-consecutive-elif
        depth_out, type_ = _get_underlying_type(
            underlying,
            allow_optional=False,
            depth=depth,
            max_depth=max_depth,
        )

    return depth_out, type_


def _get_underlying_if_optional(t: Any, pass_through: bool = False) -> Any:
    if _is_union_type(t):
        args = get_args(t)
        if len(args) == 2 and _NoneType in args:  # noqa: PLR2004
            for arg in args:
                if arg != _NoneType:
                    return arg
    elif pass_through:  # pylint: disable=confusing-consecutive-elif
        return t

    return None


def _is_union_type(t: Any) -> bool:
    # types.UnionType only exists in Python 3.10+.
    # https://docs.python.org/3/library/stdtypes.html#types-union
    import types

    origin = get_origin(t)
    return origin is types.UnionType or origin is Union  # pylint: disable=consider-alternative-union-syntax)



from __future__ import annotations

from collections.abc import Callable, Sequence  # noqa: F401
from typing import (
    TYPE_CHECKING,
    Any,
    TypedDict,
    TypeVar,
)

import numpy as np  # noqa: F401
from numpy.typing import NDArray

from ._typing_compat import ParamSpec, TypeAlias

if TYPE_CHECKING:
    from .base_potential import PhiAbstract

P = ParamSpec("P")
"""Parameter specification"""

R = TypeVar("R")
"""Return Type"""

Array: TypeAlias = NDArray["np.floating[Any]"]
ArrayLike: TypeAlias = "Sequence[float] | NDArray[np.floating[Any]]"
Float_or_ArrayLike: TypeAlias = "float | ArrayLike"
Float_or_Array: TypeAlias = "float | Array"
Phi_Signature: TypeAlias = "Callable[..., Array]"

# To bind input and output to same type
T_Float_or_Array = TypeVar("T_Float_or_Array", float, Array)


T_PhiAbstract = TypeVar("T_PhiAbstract", bound="PhiAbstract")


# Quadrature

QuadSegments_Integrals: TypeAlias = "float | list[float]"
QuadSegments_Errors: TypeAlias = "float | list[float]"
QuadSegments_Outputs: TypeAlias = "dict[str, Any] | list[dict[str, Any]]"

QuadSegments: TypeAlias = """(
    QuadSegments_Integrals
    | tuple[QuadSegments_Integrals, QuadSegments_Errors]
    | tuple[QuadSegments_Integrals, QuadSegments_Outputs]
    | tuple[QuadSegments_Integrals, QuadSegments_Errors, QuadSegments_Outputs]
)
"""


# Minimizing
class OptimizeResultInterface(TypedDict):
    """Interface to scipy.optimize.OptimizeResult."""

    x: Array
    success: bool
    status: int
    message: str
    fun: Array | float
    jac: Array
    hess: Array
    hess_inv: Any
    nfev: int
    njev: int
    nhev: int
    nit: int
    maxcv: float



"""
Classes/routines for pair potentials (:mod:`analphipy.potential`)
=================================================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import attrs
import numpy as np
from attrs import field

from ._attrs_utils import field_array_formatter, field_formatter
from ._docstrings import docfiller
from .base_potential import PhiAbstract, PhiBase

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Literal, TypeVar

    from ._typing import Array, ArrayLike, Float_or_ArrayLike, Phi_Signature


docfiller_phibase = docfiller.factory_inherit_from_parent(PhiBase)


@attrs.define(frozen=True)
@docfiller.decorate
class Generic(PhiBase):
    """
    Class to define potential using callables.

    Parameters
    ----------
    phi_func : Callable
        Function ``phi(r)``
    dphidr : Callable, optional
        Optional function ``dphidr(r)``.
    {r_min_exact}
    {phi_min_exact}
    {segments}
    """

    #: Function :math:`\phi(r)`
    phi_func: Phi_Signature

    #: Function :math:`d\phi(r)/dr`
    dphidr_func: Phi_Signature | None = None

    @docfiller_phibase()
    def phi(self, r: Float_or_ArrayLike) -> Array:
        r = np.asarray(r)
        return self.phi_func(r)

    @docfiller_phibase()
    def dphidr(self, r: Float_or_ArrayLike) -> Array:
        if self.dphidr_func is None:
            msg = "Must specify dphidr_func"
            raise ValueError(msg)
        r = np.asarray(r)
        return self.dphidr_func(r)  # pylint: disable=not-callable,useless-suppression


@attrs.define(frozen=True)
@docfiller.decorate
class Analytic(PhiBase):
    """
    Base class for defining analytic potentials.

    Parameters
    ----------
    {r_min_exact}
    {phi_min_exact}
    {segments}

    Notes
    -----
    Specific subclasses should set values for ``r_min``,
    ``phi_min``, and ``segments``, as well as
    forms for ``phi`` and ``dphidr``.

    """

    # fmt: off
    #: Position of minimum in :math:`\phi(r)`
    r_min: float = field(  # pyright: ignore[reportIncompatibleVariableOverride]
        default=np.nan, init=False, repr=field_formatter()
    )
    #: Value of ``phi`` at minimum.
    phi_min: float = field(  # pyright: ignore[reportIncompatibleVariableOverride]
        default=np.nan, init=False, repr=False
    )
    #: Integration limits.
    segments: Sequence[float] = field(  # pyright: ignore[reportIncompatibleVariableOverride]
        default=(), init=False
    )
    # fmt: on


docfiller_analytic = docfiller.factory_inherit_from_parent(Analytic)


@attrs.define(frozen=True)
@docfiller_analytic(Analytic)
class LennardJones(Analytic):
    r"""
    Lennard-Jones potential.

    .. math::

        \phi(r) = 4 \epsilon \left[ \left(\frac{{\sigma}}{{r}}\right)^{{12}} - \left(\frac{{\sigma}}{{r}}\right)^6\right]

    Parameters
    ----------
    sig : float
        Length parameter :math:`\sigma`.
    eps : float
        Energy parameter :math:`\epsilon`.

    """

    #: Length parameter :math:`\sigma`
    sig: float = 1.0
    #: Energy parameter :math:`\epsilon`
    eps: float = 1.0

    def __attrs_post_init__(self) -> None:
        self._immutable_setattrs(
            r_min=self.sig * 2.0 ** (1.0 / 6.0),
            phi_min=-self.eps,
            segments=(0.0, np.inf),
        )

    @property
    def _sigsq(self) -> float:
        return self.sig**2

    @property
    def _four_eps(self) -> float:
        return 4.0 * self.eps

    @docfiller_analytic()
    def phi(self, r: Float_or_ArrayLike) -> Array:
        r = np.array(r)
        x2: Array = self._sigsq / (r * r)
        x6: Array = x2 * x2 * x2
        return self._four_eps * x6 * (x6 - 1.0)

    @docfiller_analytic()
    def dphidr(self, r: Float_or_ArrayLike) -> Array:
        """Calculate phi and dphi (=-1/r dphi/dr) at particular r."""
        r = np.array(r)
        rinvsq = 1.0 / (r * r)

        x2: Array = self._sigsq * rinvsq
        x6: Array = x2 * x2 * x2

        return -12.0 * self._four_eps * x6 * (x6 - 0.5) / r


@attrs.define(frozen=True)
@docfiller_analytic(Analytic)
class LennardJonesNM(Analytic):
    r"""
    Generalized Lennard-Jones potential

    .. math::

        \phi(r) = \epsilon \frac{{n}}{{n-m}} \left( \frac{{n}}{{m}} \right) ^{{m / (n-m)}}
        \left[ \left(\frac{{\sigma}}{{r}}\right)^n - \left(\frac{{\sigma}}{{r}}\right)^m\right]


    Parameters
    ----------
    n, m : int
        ``n`` and ``m`` parameters to potential :math:`n, m`.
    sig : float
        Length parameter :math:`\sigma`.
    eps : float
        Energy parameter :math:`\epsilon`.


    Notes
    -----
    with parameters ``n=12`` and ``m=6``, this is equivalent to :class:`LennardJones`.

    """

    n: int = 12  #: ``n`` parameter
    m: int = 6  #: ``m`` parameter
    sig: float = 1.0  #: Length parameter
    eps: float = 1.0  #: Energy parameter

    def __attrs_post_init__(self) -> None:
        n, m = self.n, self.m

        self._immutable_setattrs(
            r_min=self.sig * (n / m) ** (1.0 / (n - m)),
            phi_min=-self.eps,
            segments=(0.0, np.inf),
        )

    @property
    def _prefac(self) -> float:
        n, m, eps = self.n, self.m, self.eps
        out = eps * (n / (n - m)) * (n / m) ** (m / (n - m))
        if isinstance(out, float):
            return out

        msg = f"Bad parameters lead to unknown prefac {out}"
        raise ValueError(msg)

    @docfiller_analytic()
    def phi(self, r: Float_or_ArrayLike) -> Array:
        r = np.array(r)

        x = self.sig / r
        return self._prefac * (x**self.n - x**self.m)

    @docfiller_analytic()
    def dphidr(self, r: Float_or_ArrayLike) -> Array:
        r = np.array(r)
        x = self.sig / r

        xn = x**self.n
        xm = x**self.m

        return -self._prefac * (self.n * xn - self.m * xm) / (r)


@attrs.define(frozen=True)
@docfiller_analytic(Analytic)
class Yukawa(Analytic):
    r"""
    Hard core Yukawa potential


    .. math::

        \phi(r) =
        \begin{{cases}}
            \infty &  r \leq \sigma \\
            -\epsilon \frac{{\sigma}}{{r}} \exp\left[-z (r/\sigma - 1) \right] & r > \sigma
        \end{{cases}}


    Parameters
    ----------
    sig : float
        Length parameters :math:`\sigma`.
    eps : float
        Energy parameter :math:`\epsilon`.
    z : float
        Interaction range parameter :math:`z`

    """

    z: float = 1.0  #: Interaction parameter
    sig: float = 1.0  #: Length parameter
    eps: float = 1.0  #: Energy parameter

    def __attrs_post_init__(self) -> None:
        self._immutable_setattrs(
            r_min=self.sig,
            phi_min=self.eps,
            segments=(0.0, self.sig, np.inf),
        )

    @docfiller_analytic()
    def phi(self, r: Float_or_ArrayLike) -> Array:
        sig, eps = self.sig, self.eps

        r = np.array(r)
        phi = np.empty_like(r)
        m = r >= sig

        phi[~m] = np.inf
        if np.any(m):
            x = r[m] / sig
            phi[m] = -eps * np.exp(-self.z * (x - 1.0)) / x
        return phi


@attrs.define(frozen=True)
@docfiller_analytic(Analytic)
class HardSphere(Analytic):
    r"""
    Hard-sphere pair potential

    .. math::

        \phi(r) =
        \begin{{cases}}
        \infty & r \leq \sigma \\
        0 & r > \sigma
        \end{{cases}}

    Parameters
    ----------
    sig: float
        Length scale parameter :math:`\sigma`

    """

    sig: float = 1.0  #: Length parameter

    def __attrs_post_init__(self) -> None:
        self._immutable_setattrs(segments=(0.0, self.sig))

    @docfiller_analytic()
    def phi(self, r: Float_or_ArrayLike) -> Array:
        r = np.array(r)

        phi = np.empty_like(r)

        m0 = r < self.sig
        phi[m0] = np.inf
        phi[~m0] = 0.0
        return phi


@attrs.define(frozen=True)
@docfiller_analytic(Analytic)
class SquareWell(Analytic):
    r"""
    Square-well pair potential


    .. math::

        \phi(r) =
        \begin{{cases}}
        \infty & r \leq \sigma \\
        \epsilon & \sigma < r \leq \lambda \sigma \\
        0 & r > \lambda \sigma
        \end{{cases}}

    Parameters
    ----------
    sig : float
        Length scale parameter :math:`\sigma`.
    eps : float
        Energy parameter :math:`\epsilon`.  Note that here, ``eps`` is the value inside the well.  So, to specify an attractive square well potential, pass a negative value for ``eps``.
    lam : float
        Width of well parameter :math:`lambda`.

    """

    sig: float = 1.0  #: Length parameter.
    eps: float = 1.0  #: Energy parameter.
    lam: float = 1.5  #: Well width parameter.

    def __attrs_post_init__(self) -> None:
        self._immutable_setattrs(
            r_min=self.sig,
            phi_min=self.eps,
            segments=(0.0, self.sig, self.sig * self.lam),
        )

    @docfiller_analytic()
    def phi(self, r: Float_or_ArrayLike) -> Array:
        sig, eps, lam = self.sig, self.eps, self.lam

        r = np.array(r)

        phi = np.empty_like(r)

        m0 = r < sig
        m2 = r >= lam * sig
        m1 = (~m0) & (~m2)

        phi[m0] = np.inf
        phi[m1] = eps
        phi[m2] = 0.0

        return phi


def _validate_bounds(self: Any, attribute: Any, bounds: Sequence[float]) -> None:  # noqa: ARG001
    if len(bounds) != 2:  # noqa: PLR2004
        msg = "length of bounds must be 2"
        raise ValueError(msg)


if TYPE_CHECKING:
    T_CubicTable = TypeVar("T_CubicTable", bound="CubicTable")


@attrs.define(frozen=True)
@docfiller.decorate
class CubicTable(PhiBase):
    """
    Cubic interpolation table potential.

    Parameters
    ----------
    {r_min_exact}
    {phi_min_exact}
    bounds : sequence of float
        the minimum and maximum values of squared pair separation `r**2` at which `phi_table` is evaluated.
    phi_table : array-like
        Values of potential evaluated on even grid of ``r**2`` values.

    segments : sequence of float, optional
        Integration segments.  Defaults to ``sqrt(bounds)``.

    phi_left, phi_right, dphi_left, dphi_right : float, optional
        Values to set for ``phi``/``-1/r dphidr`` if  (left) ``r < bounds[0]`` or (right) ``r > bounds[1]``.

    """

    #: Minimum and maximum values of squared pair separation :math:`r^2`
    bounds: Sequence[float] = field(
        validator=_validate_bounds
    )  # validator=_validate_bounds, converter=tuple)
    #: Values of potential evaluated on even grid of :math:`r^2` values.
    phi_table: ArrayLike = field(converter=np.asarray, repr=field_array_formatter())
    #: value of `phi` at left bound (`r < bounds[0]`)
    phi_left: float = field(converter=float, default=np.inf)
    #: value of `phi` at right bound (`r > bounds[1]`)
    phi_right: float = field(converter=float, default=0.0)

    #: value of `dphi` at left bound
    dphi_left: float = field(converter=float, default=np.inf)
    #: value of `dphi` at right bound
    dphi_right: float = field(converter=float, default=0.0)

    _ds: float = field(init=False, repr=False)
    _dsinv: float = field(init=False, repr=False)

    def __attrs_post_init__(self) -> None:
        assert isinstance(self.phi_table, np.ndarray)  # noqa: S101

        ds = (self.bounds[1] - self.bounds[0]) / self.size

        self._immutable_setattrs(
            _ds=ds,
            _dsinv=1.0 / ds,
        )

        if self.segments is None:  # pyright: ignore[reportUnnecessaryComparison]
            self._immutable_setattrs(segments=tuple(np.sqrt(x) for x in self.bounds))

    @classmethod
    def from_phi(
        cls: type[T_CubicTable],
        phi: Phi_Signature,
        rmin: float,
        rmax: float,
        ds: float,
        **kws: Any,
    ) -> T_CubicTable:
        """
        Create object from callable pair potential function.

        This will evaluate ``phi`` at even spacing in ``s = r**2`` space.

        Parameters
        ----------
        phi : callable
            pair potential function
        rmin : float
            minimum pair separation `r` to evaluate at.
        rmax : float
            Maximum pair separation `r` to evaluate at.
        ds : float
            spaceing in ``s = r ** 2``.
        **kws :
            Extra arguments to constructor.

        Returns
        -------
        table : CubicTable

        """
        bounds: tuple[float, float] = (rmin * rmin, rmax * rmax)

        delta = bounds[1] - bounds[0]

        size = int(delta / ds)
        ds = delta / size

        # fmt: off
        r: Array = np.sqrt(np.arange(size + 1) * ds + bounds[0])  # pyright: ignore[reportUnknownMemberType]
        # fmt: on

        phi_table = phi(r)

        return cls(bounds=bounds, phi_table=phi_table, **kws)

    def __len__(self) -> int:
        return len(self.phi_table)

    @property
    def size(self) -> int:
        """Size of the array (less 1)."""
        return len(self) - 1

    @property
    def smin(self) -> float:
        """Minimum value of `s = r**2`."""
        return self.bounds[0]

    @property
    def smax(self) -> float:
        """Maximum value of `s = r**2`."""
        return self.bounds[1]

    def phidphi(self, r: Float_or_ArrayLike) -> tuple[Array, Array]:
        """Values of `phi` and `dphi` at `r`."""
        r = np.asarray(r)

        v = np.empty_like(r)
        dv = np.empty_like(r)

        s = r * r
        left = s <= self.smin
        right = s >= self.smax
        mid = (~left) & (~right)

        v[left] = self.phi_left
        dv[right] = self.dphi_left

        v[right] = self.phi_right
        dv[right] = self.dphi_right

        if np.any(mid):
            sds = (s[mid] - self.smin) * self._dsinv
            k = sds.astype(int)
            k[k < 0] = 0

            xi = sds - k

            # fmt: off
            t: Array = np.take(self.phi_table, [k, k + 1, k + 2], mode="clip")  # pyright: ignore[reportUnknownMemberType]
            # fmt: on
            dt = np.diff(t, axis=0)
            ddt = np.diff(dt, axis=0)

            v[mid] = t[0, :] + xi * (dt[0, :] + 0.5 * (xi - 1.0) * ddt[0, :])
            dv[mid] = -2.0 * self._dsinv * (dt[0, :] + (xi - 0.5) * ddt[0, :])
        return v, dv

    @docfiller_phibase()
    def phi(self, r: Float_or_ArrayLike) -> Array:
        return self.phidphi(r)[0]

    @docfiller_phibase()
    def dphidr(self, r: Float_or_ArrayLike) -> Array:
        r = np.asarray(r)
        return -r * self.phidphi(r)[1]

    @property
    def rsq_table(self) -> Array:
        """Value of ``r**2`` where potential is defined."""
        # fmt: off
        return (self.smin + np.arange(self.size + 1) * self._ds)  # pyright: ignore[reportUnknownMemberType]
        # fmt: on

    @property
    def r_table(self) -> Array:
        """Values of ``r`` where potential is defined."""
        return np.sqrt(self.rsq_table)


if TYPE_CHECKING:
    _PHI_NAMES = Literal["lj", "nm", "sw", "hs", "yk", "LJ", "NM", "SW", "HS", "YK"]


def factory(
    potential_name: _PHI_NAMES,
    rcut: float | None = None,
    lfs: bool = False,
    cut: bool = False,
    **kws: Any,
) -> PhiAbstract:
    """
    Factory function to construct Phi object by name.

    Parameters
    ----------
    potential_name : {lj, nm, sw, hs, yk}
        Name of potential.
    rcut : float, optional
        if passed, Construct either and 'lfs' or 'cut' version of the potential.
    lfs : bool, default=False
        If True, construct a  linear force shifted potential :class:`analphipy.base_potential.PhiLFS`.
    cut : bool, default=False
        If True, construct a cut potential :class:`analphipy.base_potential.PhiCut`.

    Returns
    -------
    phi : :class:`analphipy.base_potential.PhiBase` subclass
        output potential energy class.

    See Also
    --------
    LennardJones
    LennardJonesNM
    SquareWell
    HardSphere
    Yukawa

    """
    name = potential_name.lower()

    phi: PhiAbstract

    if name == "lj":
        phi = LennardJones(**kws)

    elif name == "sw":
        phi = SquareWell(**kws)

    elif name == "nm":
        phi = LennardJonesNM(**kws)

    elif name == "yk":
        phi = Yukawa(**kws)

    elif name == "hs":
        phi = HardSphere(**kws)

    else:
        msg = f"{name} must be in {_PHI_NAMES}"
        raise ValueError(msg)

    if lfs or cut:
        if rcut is None:
            msg = "rcut cannot be None"
            raise TypeError(msg)
        if cut:
            return phi.cut(rcut=rcut)

        if lfs:
            return phi.lfs(rcut=rcut)

    return phi



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
# pylint: disable=duplicate-code
import numpy as np

from analphipy import potential


def _get_generic(p):
    return potential.Generic(phi_func=p.phi, dphidr_func=p.dphidr, segments=p.segments)


def _do_test(params, factory, kws=None, cut=False, lfs=False, phidphi=True) -> None:
    # pylint: disable=redefined-variable-type
    if kws is None:
        kws = params.asdict()

    rcut = kws.pop("rcut") if cut or lfs else None

    p = factory(**kws)
    p = _get_generic(p)

    if cut and lfs:
        msg = "Can only specify one of `cut` and `lfs`"
        raise ValueError(msg)

    if cut:
        assert isinstance(rcut, float)
        p = p.cut(rcut=rcut)
    elif lfs:
        assert isinstance(rcut, float)
        p = p.lfs(rcut=rcut)

    if phidphi:
        phi0, dphi0 = params.get_phidphi()

        phi1 = p.phi(params.r)
        dphi1 = -p.dphidr(params.r) / params.r

        np.testing.assert_allclose(dphi0, dphi1)

    else:
        phi0 = params.get_phi()
        phi1 = p.phi(params.r)

    np.testing.assert_allclose(phi0, phi1)


def test_lj(lj_params) -> None:
    _do_test(lj_params, potential.LennardJones)


def test_lj_cut(lj_cut_params) -> None:
    _do_test(lj_cut_params, potential.LennardJones, cut=True)


def test_lj_lfs(lj_lfs_params) -> None:
    _do_test(lj_lfs_params, potential.LennardJones, lfs=True)


def test_nm(nm_params) -> None:
    _do_test(nm_params, potential.LennardJonesNM)


def test_nm_cut(lj_cut_params) -> None:
    # pylint: disable=redefined-variable-type
    params = lj_cut_params

    phi0, dphi0 = params.get_phidphi()

    p = potential.LennardJonesNM(n=12, m=6, sig=params.sig, eps=params.eps).cut(
        rcut=params.rcut
    )

    p = _get_generic(p)

    phi1 = p.phi(params.r)
    dphi1 = -p.dphidr(params.r) / params.r

    np.testing.assert_allclose(phi0, phi1)
    np.testing.assert_allclose(dphi0, dphi1)


def test_nm_lfs(lj_lfs_params) -> None:
    # pylint: disable=redefined-variable-type
    params = lj_lfs_params

    phi0, dphi0 = params.get_phidphi()

    p = potential.LennardJonesNM(n=12, m=6, sig=params.sig, eps=params.eps).lfs(
        rcut=params.rcut
    )

    p = _get_generic(p)

    phi1 = p.phi(params.r)
    dphi1 = -p.dphidr(params.r) / params.r

    np.testing.assert_allclose(phi0, phi1)
    np.testing.assert_allclose(dphi0, dphi1)


def test_sw(sw_params) -> None:
    _do_test(sw_params, potential.SquareWell, phidphi=False)


def test_yk(yk_params) -> None:
    _do_test(yk_params, potential.Yukawa, kws=yk_params.asdict(), phidphi=False)



import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if sys.version_info >= (3, 10):
    from typing import Concatenate, ParamSpec, TypeAlias, TypeGuard
else:
    from typing_extensions import Concatenate, ParamSpec, TypeAlias, TypeGuard


__all__ = [
    "Concatenate",
    "ParamSpec",
    "Self",
    "TypeAlias",
    "TypeGuard",
]



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
# pylint: disable=duplicate-code
from pathlib import Path

import numpy as np
import pandas as pd

import analphipy.potential as pots
from analphipy import measures
from analphipy.norofrenkel import NoroFrenkelPair

from .utils import iter_phi_lj

data = Path(__file__).parent / "data"

nsamp = 20


def test_B2_sw() -> None:
    table = (
        pd.read_csv(data / "vir_sw_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
    )

    if len(table) > nsamp:
        table = table.sample(nsamp)
    for (sig, eps, lam), g in table.groupby(["sig", "eps", "lam"]):
        p = pots.SquareWell(sig=sig, eps=eps, lam=lam)

        a = NoroFrenkelPair(phi=p.phi, segments=p.segments, r_min=sig, phi_min=eps)

        out = a.table(np.asarray(g["beta"], dtype=np.float64), ["B2", "B2_dbeta"])

        np.testing.assert_allclose(g["Bn_2"], out["B2"])
        np.testing.assert_allclose(-g["dBn_2"], out["B2_dbeta"])


def test_B2_lj() -> None:
    table = (
        pd.read_csv(data / "vir_lj_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
        .sample(40)
    )

    for g, a in iter_phi_lj(table, nsamp):
        out = a.table(np.asarray(g["beta"], dtype=np.float64), ["B2", "B2_dbeta"])

        np.testing.assert_allclose(g["Bn_2"], out["B2"], rtol=1e-3)
        np.testing.assert_allclose(-g["dBn_2"], out["B2_dbeta"], rtol=1e-3)


def test_B2_nm() -> None:
    table = (
        pd.read_csv(data / "vir_lj_nm_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
    )

    if len(table) > nsamp:
        table = table.sample(nsamp)
    sig = eps = 1.0

    for (n_exp, m_exp), g in table.groupby(["n_exp", "m_exp"]):
        p = pots.LennardJonesNM(n=n_exp, m=m_exp, sig=sig, eps=eps)

        a = NoroFrenkelPair.from_phi(
            phi=p.phi, segments=p.segments, r_min=sig, bounds=[0.5, 1.5]
        )

        out = a.table(np.asarray(g["beta"], dtype=np.float64), ["B2", "B2_dbeta"])

        np.testing.assert_allclose(g["Bn_2"], out["B2"], rtol=1e-3)
        np.testing.assert_allclose(-g["dBn_2"], out["B2_dbeta"], rtol=1e-3)


def test_B2_yk() -> None:
    table = (
        pd.read_csv(data / "vir_yukawa_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
    )

    if len(table) > nsamp:
        table = table.sample(nsamp)
    sig = eps = 1.0

    for (z_yukawa), g in table.groupby("z_yukawa"):
        assert isinstance(z_yukawa, float)
        p = pots.Yukawa(z=z_yukawa, sig=sig, eps=eps)

        a = NoroFrenkelPair(phi=p.phi, segments=p.segments, r_min=sig, phi_min=-1.0)

        out = a.table(np.asarray(g["beta"], dtype=np.float64), ["B2", "B2_dbeta"])

        np.testing.assert_allclose(g["Bn_2"], out["B2"], rtol=1e-6)
        np.testing.assert_allclose(-g["dBn_2"], out["B2_dbeta"], rtol=1e-6)


def test_B2_hs() -> None:
    rng = np.random.default_rng()
    sig = rng.random()

    p = pots.HardSphere(sig=sig)

    B2a = 2 * np.pi / 3.0 * sig**3

    B2b = measures.secondvirial(p.phi, beta=1.0, segments=[0.0, sig])

    assert isinstance(B2b, float)

    np.testing.assert_allclose(B2a, B2b)

    B2_dbeta = measures.secondvirial_dbeta(p.phi, beta=1.0, segments=[0.0, sig])

    assert isinstance(B2_dbeta, float)

    np.testing.assert_allclose(0.0, B2_dbeta)



"""
Base classes (:mod:`analphipy.base_potential`)
==============================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

import attrs
import numpy as np
from attrs import field

from ._attrs_utils import field_formatter
from ._docstrings import docfiller
from .measures import Measures
from .norofrenkel import NoroFrenkelPair
from .utils import minimize_phi

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any, Literal

    from ._typing import Array, Float_or_ArrayLike
    from ._typing_compat import Self


# * attrs utilities
def segments_converter(segments: Sequence[Any]) -> tuple[float, ...]:
    """Make sure segments are in correct format."""
    return tuple(float(x) for x in segments)


@attrs.define(frozen=True, kw_only=True)
@docfiller.decorate
class PhiAbstract:
    """
    Abstract class from which base classes inherit.

    Parameters
    ----------
    {r_min_exact}
    {phi_min_exact}
    {segments}

    """

    # pylint: disable=redundant-returns-doc,no-self-use

    #: Position of minimum in :math:`\phi(r)`
    r_min: float | None = attrs.field(
        default=None,
        converter=attrs.converters.optional(float),
        repr=field_formatter(),
    )

    #: Value of ``phi`` at minimum.
    phi_min: float | None = attrs.field(
        default=None, converter=attrs.converters.optional(float), repr=field_formatter()
    )

    #: Integration limits
    segments: Sequence[float] = attrs.field(
        default=None,
        converter=attrs.converters.optional(segments_converter),
    )

    def asdict(self) -> dict[str, Any]:
        """Convert object to dictionary."""
        return attrs.asdict(self, filter=self._get_smart_filter())

    def new_like(self, **kws: Any) -> Self:
        """
        Create a new object with optional parameters.

        Parameters
        ----------
        **kws
            `attribute`, `value` pairs.

        """
        return attrs.evolve(self, **kws)

    def assign(self, **kws: Any) -> Self:
        """Alias to :meth:`new_like`."""
        return self.new_like(**kws)

    def _immutable_setattrs(self, **kws: Any) -> None:
        """
        Set attributes of frozen attrs class.

        This should only be used to set 'derived' attributes on creation.

        Parameters
        ----------
        **kws :
            `key`, `value` pairs to be set.  `key` should be the name of the attribute to be set,
            and `value` is the value to set `key` to.

        Examples
        --------
        >>> @attrs.frozen
        ... class Derived(PhiAbstract):
        ...     _derived = attrs.field(init=False)
        ...
        ...     def __attrs_post_init__(self):
        ...         self._immutable_setattrs(_derived=self.r_min + 10)

        >>> x = Derived(r_min=5)
        >>> x._derived
        15.0

        """
        for key, value in kws.items():
            object.__setattr__(self, key, value)  # noqa: PLC2801

    def _get_smart_filter(
        self,
        include: Sequence[str] | None = None,
        exclude: Sequence[str] | None = None,
        exclude_private: bool = True,
        exclude_no_init: bool = True,
    ) -> Callable[..., Any]:  # pragma: no cover
        """
        Create a filter to include exclude names.

        Parameters
        ----------
        include : sequence of str
            Names to include
        exclude : sequence of str
            names to exclude
        exclude_private : bool, default=True
            If True, exclude any names starting with '_'
        exclude_no_init : bool, default=True
            If True, exclude any fields defined with ``field(init=False)``.

        Notes
        -----
        Precedence is in order
        `include, exclude, exclude_private exclude_no_init`.
        That is, if a name is in include and exclude and is private/no_init,
        it will be included

        """
        fields = attrs.fields(type(self))

        if include is None:
            include = []
        elif isinstance(include, str):
            include = [include]

        if exclude is None:
            exclude = []
        elif isinstance(exclude, str):
            exclude = [exclude]

        includes: list[Any] = []
        for f in fields:
            if f.name in include:
                includes.append(f)

            elif (
                f.name in exclude
                or (exclude_private and f.name.startswith("_"))
                or (exclude_no_init and not f.init)
            ):
                pass

            else:
                includes.append(f)
        return attrs.filters.include(*includes)

    def phi(self, r: Float_or_ArrayLike) -> Array:
        """
        Pair potential.

        Parameters
        ----------
        r : array-like
            pair separation

        Returns
        -------
        phi : ndarray
            Evaluated pair potential.

        """
        msg = "Must implement in subclass"
        raise NotImplementedError(msg)

    def dphidr(self, r: Float_or_ArrayLike) -> Array:
        r"""
        Derivative of pair potential.

        This returns the value of :math:`d \phi(r) / dr`

        Parameters
        ----------
        r : array-like
            pair separation

        Returns
        -------
        dphidr : ndarray
            Pair potential values.

        """
        msg = "Must implement in subclass"
        raise NotImplementedError(msg)

    def minimize(
        self,
        r0: float | Literal["mean"],
        bounds: tuple[float, float] | Literal["segments"] | None = None,
        **kws: Any,
    ) -> tuple[float, float, Any]:
        """
        Determine position `r` where ``phi`` is minimized.

        Parameters
        ----------
        r0 : float
            Guess for position of minimum.
            If value is `'mean'`, then use ``r0 = mean(bounds)``.

        bounds : tuple of float, {'segments'}, optional
            Bounds for minimization search.
            If tuple, should be of form ``bounds=(lower_bound, upper_bound)``.
            If `'segments`, then ``bounds=(segments[0], segments[-1])``.
            If None, no bounds used.
        **kws :
            Extra arguments to :func:`analphipy.utils.minimize_phi`

        Returns
        -------
        rmin : float
            ``phi(rmin)`` is found location of minimum.
        phimin : float
            Value of ``phi(rmin)``, i.e., the value of ``phi`` at the minimum
        output : object
            Output class from :func:`scipy.optimize.minimize`.

        See Also
        --------
        ~analphipy.utils.minimize_phi

        """
        if bounds == "segments":
            if self.segments is not None:  # pyright: ignore[reportUnnecessaryComparison]
                bounds = (self.segments[0], self.segments[-1])
            else:
                bounds = None  # pragma: no cover

        if r0 == "mean":
            if bounds is not None:
                if not isinstance(bounds, tuple):  # pyright: ignore[reportUnnecessaryIsInstance]
                    msg = "bounds must be a tuple"
                    raise TypeError(msg)
                r0 = cast("float", np.mean(bounds))
            else:
                msg = 'must specify bounds with r0="mean"'
                raise ValueError(msg)

        return minimize_phi(self.phi, r0=r0, bounds=bounds, **kws)

    def assign_min_numeric(
        self,
        r0: float | Literal["mean"],
        bounds: tuple[float, float] | Literal["segments"] | None = None,
        **kws: Any,
    ) -> Self:
        """
        Create new object with minima set by numerical minimization.

        call :meth:`minimize`
        """
        r_min, phi_min, _ = self.minimize(r0=r0, bounds=bounds, **kws)
        return self.new_like(r_min=r_min, phi_min=phi_min)

    def to_nf(self, **kws: Any) -> NoroFrenkelPair:
        """
        Create a :class:`analphipy.norofrenkel.NoroFrenkelPair` object.

        Parameters
        ----------
        **kws :
            Extra arguments to :class:`analphipy.norofrenkel.NoroFrenkelPair` constructor.
            parameters `phi`, `semgnets`, `r_min', `phi_min` default to values from `self`.

        Returns
        -------
        nf : :class:`analphipy.norofrenkel.NoroFrenkelPair`

        """
        if self.r_min is None:
            msg = "must set `self.r_min` to use NoroFrenkel"
            raise ValueError(msg)

        for k in ("phi", "segments", "r_min", "phi_min"):
            if k not in kws:
                kws[k] = getattr(self, k)

        return NoroFrenkelPair(**kws)

    def to_measures(self, **kws: Any) -> Measures:
        """
        Create a :class:`analphipy.measures.Measures` object.

        Parameters
        ----------
        **kws :
            Extra arguments to :class:`analphipy.measures.Measures` constructor.
            parameters `phi`, `semgnets` default to values from `self`.

        Returns
        -------
        nf : :class:`analphipy.measures.Measures`

        """
        for k in ("phi", "segments"):
            if k not in kws:
                kws[k] = getattr(self, k)

        return Measures(**kws)


docfiller_phiabstract = docfiller.factory_inherit_from_parent(PhiAbstract)


@attrs.define(frozen=True)
@docfiller.inherit(PhiAbstract)
class PhiCutBase(PhiAbstract):
    r"""
    Base Class for creating cut potential from base potential


    .. math::

        \phi_{{\rm cut}}(r) =
            \begin{{cases}}
                \phi(r) + {{\text _vcorrect}}(r) & r < r_{{\rm cut}} \\
                0 & r_{{\rm cut}} \leq r
            \end{{cases}}

        \frac{{d \phi_{{\rm cut}}(r)}}{{d r}} =
            \begin{{cases}}
                \frac{{d \phi(r)}}{{d r}} + {{\text _dvdrcorrect}}(r) & r < r_{{\rm cut}} \\
                0 & r_{{\rm cut}} \leq r
            \end{{cases}}

    So, for example, for just cut, `_vcorrect(r) = -v(rcut)`, `dvrcut=0`,
    and for lfs, `_vcorrect(r) = -v(rcut) - dv(rcut)/dr (r - rcut)`
    `_dvcorrect(r) = ...`


    Parameters
    ----------
    phi_base : :class:`analphipy.base_potential.PhiAbstract`
    rcut : float
        Position to cut the potential.


    """

    #: Instance of (sub)class of :class:`analphipy.base_potential.PhiAbstract`
    phi_base: PhiBase
    #: Position to cut the potential
    rcut: float = field(converter=float)
    #: Integration limits
    segments: Sequence[float] = field(init=False, repr=False)  # pyright: ignore[reportGeneralTypeIssues, reportIncompatibleVariableOverride]

    def __attrs_post_init__(self) -> None:
        if self.phi_base.segments is None:  # pyright: ignore[reportUnnecessaryComparison]
            msg = "must specify segments"
            raise ValueError(msg)  # pragma: no cover
        self._immutable_setattrs(
            segments=(
                *tuple(x for x in self.phi_base.segments if x < self.rcut),
                self.rcut,
            )
        )

    @docfiller_phiabstract()
    def phi(self, r: Float_or_ArrayLike) -> Array:  # noqa: D102
        r = np.asarray(r)
        v = np.empty_like(r)

        left = r <= self.rcut
        right = ~left

        v[right] = 0.0

        if np.any(left):
            v[left] = self.phi_base.phi(r[left]) + self._vcorrect(r[left])
        return v

    @docfiller_phiabstract()
    def dphidr(self, r: Float_or_ArrayLike) -> Array:  # noqa: D102
        r = np.asarray(r)
        dvdr = np.empty_like(r)

        left = r <= self.rcut
        right = ~left

        dvdr[right] = 0.0
        dvdr[left] = self.phi_base.dphidr(r[left]) + self._dvdrcorrect(r[left])

        return dvdr

    def _vcorrect(self, r: Array) -> Array:
        raise NotImplementedError

    def _dvdrcorrect(self, r: Array) -> Array:
        raise NotImplementedError


@attrs.frozen
@docfiller.inherit(PhiCutBase)
class PhiCut(PhiCutBase):
    r"""
    Pair potential cut at position ``r_cut``.

    .. math::

        \phi_{{\rm cut}}(r) =
            \begin{{cases}}
                \phi(r) - \phi(r_{{\rm cut}})  & r < r_{{\rm cut}} \\
                0 & r_{{\rm cut}} \leq r
            \end{{cases}}

    Parameters
    ----------
    phi_base : :class:`analphipy.base_potential.PhiBase` instance
        Potential class to perform cut on.
    rcut : float
        Where to 'cut' the potential.

    """

    _vcut: float = field(init=False, repr=False)

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        self._immutable_setattrs(_vcut=self.phi_base.phi(self.rcut))

    def _vcorrect(self, r: Array) -> Array:
        return -cast("Array", self._vcut)

    def _dvdrcorrect(self, r: Array) -> Array:
        return np.array(0.0)


@attrs.frozen
@docfiller.inherit(PhiCutBase)
class PhiLFS(PhiCutBase):
    r"""
    Pair potential cut and linear force shifted at ``r_cut``.

    .. math::

        \phi_{{\rm lfs}}(r) =
            \begin{{cases}}
                \phi(r)
                - \left( \frac{{d \phi}}{{d r}} \right)_{{\rm cut}} (r - r_{{\rm cut}})
                - \phi(r_{{\rm cut}}) & r < r_{{\rm cut}} \\
                0 & r_{{\rm cut}} < r
            \end{{cases}}
    """

    _vcut: float = field(init=False, repr=False)
    _dvdrcut: float = field(init=False, repr=False)

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        self._immutable_setattrs(
            _vcut=self.phi_base.phi(self.rcut), _dvdrcut=self.phi_base.dphidr(self.rcut)
        )

    def _vcorrect(self, r: Array) -> Array:
        return -(self._vcut + self._dvdrcut * (r - self.rcut))

    def _dvdrcorrect(self, r: Array) -> Array:
        out = -self._dvdrcut
        return cast("Array", out)


@attrs.frozen
@docfiller.inherit(PhiAbstract)
class PhiBase(PhiAbstract):
    """Base class for potentials."""

    def cut(self, rcut: float) -> PhiCut:
        """Create a :class:`PhiCut` potential."""
        return PhiCut(phi_base=self, rcut=rcut)

    def lfs(self, rcut: float) -> PhiLFS:
        """Create a :class:`PhiLFS` potential."""
        return PhiLFS(phi_base=self, rcut=rcut)



"""
Routines to calculate measures of pair potentials (:mod:`analphipy.measures`)
=============================================================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np
from module_utilities import cached

from ._docstrings import docfiller
from .utils import TWO_PI, add_quad_kws, combine_segmets, quad_segments

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence
    from typing import Any

    from ._typing import (
        Array,
        ArrayLike,
        Float_or_Array,
        Float_or_ArrayLike,
        Phi_Signature,
        QuadSegments,
    )
    from .base_potential import PhiAbstract


__all__ = [
    "Measures",
    "diverg_js_cont",
    "diverg_kl_cont",
    "secondvirial",
    "secondvirial_dbeta",
    "secondvirial_sw",
]


@docfiller.decorate
def secondvirial(
    phi: Phi_Signature,
    beta: float,
    segments: ArrayLike,
    err: bool = False,
    full_output: bool = False,
    **kws: Any,
) -> QuadSegments:
    r"""
    Calculate the second virial coefficient.

    .. math::

        B_2(\beta) = -\int 2\pi r^2 dr \left(\exp(-\beta \phi(r)) - 1\right)

    Parameters
    ----------
    {phi}
    {beta}
    {segments}
    {err}
    {full_output}
    **kws
        Extra arguments to :func:`analphipy.utils.quad_segments`

    Returns
    -------
    B2 : float
        Value of second virial coefficient.
    {error_summed}
    {full_output_summed}

    See Also
    --------
    ~analphipy.utils.quad_segments

    """

    def integrand(r: Float_or_Array) -> Array:
        out: Array = TWO_PI * r**2 * (1 - np.exp(-beta * phi(r)))
        return out

    return quad_segments(
        integrand,
        segments=segments,
        sum_integrals=True,
        sum_errors=True,
        err=err,
        full_output=full_output,
        **kws,
    )


@docfiller.decorate
def secondvirial_dbeta(
    phi: Phi_Signature,
    beta: float,
    segments: ArrayLike,
    err: bool = False,
    full_output: bool = False,
    **kws: Any,
) -> QuadSegments:
    r"""
    ``beta`` derivative of second virial coefficient.

    .. math::

        \frac{{d B_2}}{{d \beta}} = \int 2\pi r^2 dr \phi(r) \exp(-\beta \phi(r))

    Parameters
    ----------
    {phi}
    {beta}
    {segments}
    {err}
    {full_output}

    Returns
    -------
    dB2dbeta : float
        Value of derivative.
    {error_summed}
    {full_output_summed}


    """

    def integrand(r: Float_or_Array) -> Array:
        v = phi(r)
        out = np.array(0.0) if np.isinf(v) else TWO_PI * r**2 * v * np.exp(-beta * v)
        return cast("Array", out)

    return quad_segments(
        integrand,
        segments=segments,
        sum_integrals=True,
        sum_errors=True,
        err=err,
        full_output=full_output,
        **kws,
    )


@docfiller.decorate
def secondvirial_sw(beta: float, sig: float, eps: float, lam: float) -> float:
    r"""
    Second virial coefficient for a square well (SW) fluid. Note that this assumes that
    the SW fluid is defined by the potential:

    .. math::

        \phi(r) =
        \begin{{cases}}
            \infty & r \leq \sigma \\
            \epsilon & \sigma < r \leq \lambda \sigma  \\
            0 & \lambda \sigma < r
        \end{{cases}}


    Parameters
    ----------
    {beta}
    sig : float
        Length parameter :math:`\sigma`.
    eps : float
        Energy parameter :math:`\epsilon`.
    lam : float
        Well width parameter :math:`\lambda`.

    Returns
    -------
    B2 : float
        Value of second virial coefficient.

    """
    out: float = (
        TWO_PI / 3.0 * sig**3 * (1.0 + (1 - np.exp(-beta * eps)) * (lam**3 - 1.0))
    )
    return out


def diverg_kl_integrand(
    p: Float_or_ArrayLike,
    q: Float_or_ArrayLike,
    volume: Float_or_ArrayLike | None = None,
) -> Array:
    p, q = np.asarray(p), np.asarray(q)

    out = np.empty_like(p)

    zero = p == 0.0  # pylint: disable=use-implicit-booleaness-not-comparison-to-zero
    hero = ~zero

    out[zero] = 0.0
    out[hero] = p[hero] * np.log(p[hero] / q[hero])

    if volume is not None:
        out *= np.asarray(volume)

    return out


@docfiller(summary="Calculate discrete Kullback-Leibler divergence")
def diverg_kl_disc(
    p: Float_or_ArrayLike, q: Float_or_ArrayLike, axis: int | None = None
) -> Float_or_Array:
    """
    {summary}.

    Parameters
    ----------
    p, q : array-like
        Probabilities to consider

    Returns
    -------
    results : float or ndarray
        Value of divergence.

    References
    ----------
    {kl_link}

    """
    p, q = np.asarray(p), np.asarray(q)
    diverg = diverg_kl_integrand(p, q)
    # fmt: off
    out: Float_or_Array = diverg.sum(axis=axis)  # pyright: ignore[reportUnknownMemberType]
    # fmt: on
    return out


def _check_volume_func(
    volume: str | Callable[[Float_or_Array], Float_or_Array] | None = None,
) -> Callable[[Float_or_Array], Float_or_Array]:
    if volume is None:
        volume = "1d"
    if isinstance(volume, str):
        if volume == "1d":
            return lambda x: 1.0  # noqa: ARG005

        if volume == "2d":
            return lambda x: 2.0 * np.pi * x

        if volume == "3d":
            return lambda x: 4 * np.pi * x**2

        msg = "unknown dimension"
        raise ValueError(msg)

    if not callable(volume):
        msg = "volume should be str or callable"
        raise TypeError(msg)

    return volume


@docfiller(
    summary="Calculate continuous Kullback-Leibler divergence for continuous pdf"
)
def diverg_kl_cont(
    p: Callable[[Float_or_Array], Float_or_Array],
    q: Callable[[Float_or_Array], Float_or_Array],
    segments: ArrayLike,
    segments_q: ArrayLike | None = None,
    volume: str | Callable[[Float_or_Array], Float_or_Array] | None = None,
    err: bool = False,
    full_output: bool = False,
    **kws: Any,
) -> QuadSegments:
    """
    {summary}.

    Parameters
    ----------
    p, q : callable
        Probabilities to consider
    {volume_int_func}
    {segments}
    segments_q : list, optional
        if supplied, build total segments by combining segments and segments_q
    {err}
    {full_output}

    Returns
    -------
    result : float or ndarray
        value of divergence
    {error_summed}
    {full_output_summed}

    References
    ----------
    {kl_link}

    """
    volume_callable = _check_volume_func(volume)

    if segments_q is not None:
        segments = combine_segmets(segments, segments_q)

    def func(x: Float_or_Array) -> Array:
        return diverg_kl_integrand(p=p(x), q=q(x), volume=volume_callable(x))

    return quad_segments(
        func,
        segments=segments,
        sum_integrals=True,
        sum_errors=True,
        err=err,
        full_output=full_output,
        **kws,
    )


# @doc_inherit(diverg_kl_disc, style="numpy_with_merge")
@docfiller(diverg_kl_disc, summary="Discrete Jensen-Shannon divergence")
def diverg_js_disc(
    p: Float_or_ArrayLike, q: Float_or_ArrayLike, axis: int | None = None
) -> Float_or_Array:
    p, q = np.asarray(p), np.asarray(q)

    m = 0.5 * (p + q)

    return 0.5 * (diverg_kl_disc(p, m, axis=axis) + diverg_kl_disc(q, m, axis=axis))


def diverg_js_integrand(
    p: Float_or_ArrayLike,
    q: Float_or_ArrayLike,
    volume: Float_or_ArrayLike | None = None,
) -> Array:
    p = np.asarray(p)
    q = np.asarray(q)

    m = 0.5 * (p + q)

    out = 0.5 * (diverg_kl_integrand(p, m) + diverg_kl_integrand(q, m))

    if volume is not None:
        out *= np.asarray(volume)

    return out


@docfiller(diverg_kl_cont, summary="Continuous Jensen-Shannon divergence")
def diverg_js_cont(
    p: Callable[[Float_or_Array], Float_or_Array],
    q: Callable[[Float_or_Array], Float_or_Array],
    segments: ArrayLike,
    segments_q: ArrayLike | None = None,
    volume: str | Callable[[Float_or_Array], Float_or_Array] | None = None,
    err: bool = False,
    full_output: bool = False,
    **kws: Any,
) -> QuadSegments:
    """Calculate Jensen-Shannon divergence for continuous functions."""
    volume_callable = _check_volume_func(volume)

    if segments_q is not None:
        segments = combine_segmets(segments, segments_q)

    def func(x: Float_or_Array) -> Array:
        return diverg_js_integrand(p(x), q(x), volume_callable(x))

    return quad_segments(
        func,
        segments=segments,
        sum_integrals=True,
        sum_errors=True,
        err=err,
        full_output=full_output,
        **kws,
    )


@docfiller.decorate
class Measures:
    """
    Convenience class for calculating measures.

    Parameters
    ----------
    {phi}
    {segments}
    {quad_kws}

    """

    def __init__(
        self,
        phi: Phi_Signature,
        segments: Sequence[float],
        quad_kws: Mapping[str, Any] | None = None,
    ) -> None:
        self.phi = phi
        self.segments = segments
        if quad_kws is None:
            quad_kws = {}
        self.quad_kws = quad_kws
        self._cache: dict[str, Any] = {}

    @cached.meth
    @add_quad_kws
    @docfiller.decorate
    def secondvirial(  # pylint: disable=missing-type-doc
        self, /, beta: float, err: bool = False, full_output: bool = False, **kws: Any
    ) -> QuadSegments:
        """
        Calculate second virial coefficient.

        Parameters
        ----------
        {beta}
        {err}
        {full_output}
        **kws
            Extra arguments to :func:`analphipy.utils.quad_segments`

        Returns
        -------
        B2 : float
            Value of second virial coefficient.
        {error_summed}
        {full_output_summed}

        See Also
        --------
        ~analphipy.measures.secondvirial

        """
        return secondvirial(
            phi=self.phi,
            beta=beta,
            segments=self.segments,
            err=err,
            full_output=full_output,
            **kws,
        )

    @cached.meth
    @add_quad_kws
    @docfiller.decorate
    def secondvirial_dbeta(
        self, /, beta: float, err: bool = False, full_output: bool = False, **kws: Any
    ) -> QuadSegments:
        """
        Calculate ``beta`` derivative of second virial coefficient.

        Parameters
        ----------
        {beta}
        {err}
        {full_output}

        Returns
        -------
        dB2dbeta : float
            Value of derivative.
        {error_summed}
        {full_output_summed}

        See Also
        --------
        ~analphipy.measures.secondvirial_dbeta

        """
        return secondvirial_dbeta(
            phi=self.phi,
            beta=beta,
            segments=self.segments,
            err=err,
            full_output=full_output,
            **kws,
        )

    @docfiller.decorate
    @add_quad_kws
    def boltz_diverg_js(  # pylint: disable=missing-type-doc
        self,
        /,
        other: PhiAbstract,
        beta: float,
        beta_other: float | None = None,
        volume: str | Callable[[Float_or_Array], Float_or_Array] = "3d",
        err: bool = False,
        full_output: bool = False,
        **kws: Any,
    ) -> QuadSegments:
        r"""
        Jensen-Shannon divergence of the Boltzmann factors of two potentials.

        The Boltzmann factors are defined as:

        .. math::

            B(r; \beta, \phi) = \exp(-\beta \phi(r))


        Parameters
        ----------
        other : :class:`analphipy.base_potential.PhiAbstract`
            Class wrapping other potential to compare `self` to.
        {beta}
        beta_other : float, optional
            beta value to evaluate other Boltzmann factor at.
        {volume_int_func}

        See Also
        --------
        ~analphipy.measures.diverg_js_cont

        References
        ----------
        `See here for more info <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence#Symmetrised_divergence>`

        """
        if beta_other is None:
            beta_other = beta

        def p_func(x: Float_or_Array) -> Array:
            return cast("Array", np.exp(-beta * self.phi(x)))

        def q_func(x: Float_or_Array) -> Array:
            return cast("Array", np.exp(-beta_other * other.phi(x)))

        return diverg_js_cont(
            p=p_func,
            q=q_func,
            segments=self.segments,
            segments_q=other.segments,
            volume=volume,
            err=err,
            full_output=full_output,
            **kws,
        )

    def mayer_diverg_js(
        self,
        other: PhiAbstract,
        beta: float,
        beta_other: float | None = None,
        volume: str | Callable[[Float_or_Array], Float_or_Array] = "3d",
        err: bool = False,
        full_output: bool = False,
        **kws: Any,
    ) -> QuadSegments:
        r"""
        Jensen-Shannon divergence of the Mayer f-functions of two potentials.

        The Mayer f-function is defined as:

        .. math::

            f(r; \beta, \phi) = \exp(-beta \phi(r)) - 1


        Parameters
        ----------
        other : :class:`analphipy.base_potential.PhiAbstract`
            Class wrapping other potential to compare `self` to.
        {beta}
        beta_other : float, optional
            beta value to evaluate other Boltzmann factor at.
        {volume_int_func}


        See Also
        --------
        ~analphipy.measures.diverg_js_cont


        References
        ----------
        `See here for more info <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence#Symmetrised_divergence>`

        """
        if beta_other is None:
            beta_other = beta

        def p_func(x: Float_or_Array) -> Array:
            out: Array = np.exp(-beta * self.phi(x)) - 1.0
            return out

        def q_func(x: Float_or_Array) -> Array:
            out: Array = np.exp(-beta_other * other.phi(x)) - 1.0
            return out

        return diverg_js_cont(
            p=p_func,
            q=q_func,
            segments=self.segments,
            segments_q=other.segments,
            volume=volume,
            err=err,
            full_output=full_output,
            **kws,
        )



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
# pylint: disable=unnecessary-lambda-assignment
"""Simple tests for js divergence..."""

import numpy as np
import pytest

import analphipy
from analphipy.measures import diverg_js_cont


@pytest.mark.parametrize("beta", [0.5, 1.0, 2.0])
@pytest.mark.parametrize("beta_other", [0.5, 1.0])
def test_boltz_js_lj(beta, beta_other) -> None:
    p0 = analphipy.potential.LennardJones(sig=1.0, eps=1.0)
    p1 = analphipy.potential.LennardJonesNM(n=14, m=7, sig=1.0, eps=1)

    v0 = p0.to_measures().boltz_diverg_js(p1, beta=beta, beta_other=beta_other)

    f0 = lambda x: np.exp(-beta * p0.phi(x))
    f1 = lambda x: np.exp(-beta_other * p1.phi(x))

    v1 = diverg_js_cont(
        f0, f1, segments=p0.segments, segments_q=p1.segments, volume="3d"
    )

    np.testing.assert_allclose(v0, v1)  # type: ignore[arg-type]



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
# pylint: disable=duplicate-code
import numpy as np
import pytest

from analphipy import potential as pots
from analphipy.base_potential import PhiAbstract


def test_asdict() -> None:
    # pylint: disable=redefined-variable-type
    p_lj = pots.LennardJones(sig=1.5, eps=1.5)

    p = p_lj

    assert p.asdict() == {"sig": 1.5, "eps": 1.5}

    p_new = p.new_like(sig=1.0, eps=1.0)

    assert p_new.asdict() == {"sig": 1.0, "eps": 1.0}

    assert p.assign(sig=2.0).asdict() == {"sig": 2.0, "eps": 1.5}

    pa = PhiAbstract()

    with pytest.raises(NotImplementedError):
        pa.phi(1.0)

    with pytest.raises(NotImplementedError):
        pa.dphidr(1.0)

    from analphipy.base_potential import PhiCutBase

    p = PhiCutBase(p_lj, rcut=2.5)  # type: ignore[assignment]

    with pytest.raises(NotImplementedError):
        p.phi(1.0)

    with pytest.raises(NotImplementedError):
        p.dphidr(1.0)

    p_lj = pots.LennardJones()

    p_lfs = p_lj.lfs(rcut=2.5).assign_min_numeric(1.1, bounds=(0.5, 1.5))
    np.testing.assert_allclose(p_lfs.r_min, 1.123148919)  # type: ignore[arg-type]

    p_lfs = p_lj.lfs(rcut=2.5).assign_min_numeric(1.1, bounds="segments")
    np.testing.assert_allclose(p_lfs.r_min, 1.123148919)  # type: ignore[arg-type]

    p_lfs = p_lj.lfs(rcut=2.5).assign_min_numeric(r0="mean", bounds=(0.5, 1.5))
    np.testing.assert_allclose(p_lfs.r_min, 1.123148919, atol=1e-5)  # type: ignore[arg-type]

    with pytest.raises(ValueError):
        p_lj.lfs(rcut=2.5).assign_min_numeric(r0="mean", bounds=None)

    with pytest.raises(ValueError):
        p_lj.lfs(rcut=2.5).to_nf()


def _do_test(params, factory, kws=None, cut=False, lfs=False, phidphi=True) -> None:
    if kws is None:
        kws = params.asdict()

    rcut = kws.pop("rcut") if cut or lfs else None

    p = factory(**kws)

    if cut:
        p = p.cut(rcut=rcut)
    if lfs:
        p = p.lfs(rcut=rcut)

    if phidphi:
        phi0, dphi0 = params.get_phidphi()

        phi1 = p.phi(params.r)
        dphi1 = -p.dphidr(params.r) / params.r

        np.testing.assert_allclose(dphi0, dphi1)

    else:
        phi0 = params.get_phi()
        phi1 = p.phi(params.r)

    np.testing.assert_allclose(phi0, phi1)


def test_lj(lj_params) -> None:
    _do_test(lj_params, pots.LennardJones)


def test_lj_cut(lj_cut_params) -> None:
    _do_test(lj_cut_params, pots.LennardJones, cut=True)


def test_lj_lfs(lj_lfs_params) -> None:
    _do_test(lj_lfs_params, pots.LennardJones, lfs=True)


def test_nm(nm_params) -> None:
    _do_test(nm_params, pots.LennardJonesNM)


def test_nm_cut(lj_cut_params) -> None:
    params = lj_cut_params

    phi0, dphi0 = params.get_phidphi()

    p = pots.LennardJonesNM(n=12, m=6, sig=params.sig, eps=params.eps).cut(
        rcut=params.rcut
    )

    phi1 = p.phi(params.r)
    dphi1 = -p.dphidr(params.r) / params.r

    np.testing.assert_allclose(phi0, phi1)
    np.testing.assert_allclose(dphi0, dphi1)


def test_nm_lfs(lj_lfs_params) -> None:
    params = lj_lfs_params

    phi0, dphi0 = params.get_phidphi()

    p = pots.LennardJonesNM(n=12, m=6, sig=params.sig, eps=params.eps).lfs(
        rcut=params.rcut
    )

    phi1 = p.phi(params.r)
    dphi1 = -p.dphidr(params.r) / params.r

    np.testing.assert_allclose(phi0, phi1)
    np.testing.assert_allclose(dphi0, dphi1)


def test_sw(sw_params) -> None:
    _do_test(sw_params, pots.SquareWell, phidphi=False)


def test_yk(yk_params) -> None:
    _do_test(yk_params, pots.Yukawa, kws=yk_params.asdict(), phidphi=False)


def test_hs(hs_params) -> None:
    _do_test(hs_params, pots.HardSphere, phidphi=False)



from __future__ import annotations

from typing import TYPE_CHECKING

import attrs
import numpy as np

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, TypeVar

    T = TypeVar("T", bound=Any)
    R = TypeVar("R")


def optional_converter(converter: Callable[[T], R]) -> Callable[[T], T | R]:
    """Create a converter which can pass through None."""

    def wrapped(value: T) -> T | R:  # pragma: no cover
        if value in {None, attrs.NOTHING}:
            return value  # pyright: ignore[reportReturnType]
        return converter(value)

    return wrapped


def field_formatter(fmt: str = "{:.5g}") -> Callable[[Any], str]:
    """Formatter for attrs field."""

    @optional_converter
    def wrapped(value: Any) -> str:
        return fmt.format(value)

    return wrapped


def field_array_formatter(threshold: int = 3, **kws: Any) -> Callable[[Any], str]:
    """Formatter for numpy array field."""

    @optional_converter
    def wrapped(value: Any) -> str:
        with np.printoptions(threshold=threshold, **kws):
            return str(value)

    return wrapped



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
# pylint: disable=duplicate-code
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import analphipy.potential as pots
from analphipy.norofrenkel import NoroFrenkelPair

from .utils import iter_phi_lj

data = Path(__file__).parent / "data"

nsamp = 20


def test_simple() -> None:
    p = pots.LennardJones()

    n = p.to_nf()

    np.testing.assert_allclose(n.secondvirial(1.0), n.secondvirial_sw(1.0))  # type: ignore[arg-type]

    np.testing.assert_allclose(n.secondvirial(1.0), n.B2_sw(1.0))  # type: ignore[arg-type]

    with pytest.raises(ValueError):
        n.lam(beta=1.0, err=True)

    with pytest.raises(ValueError):
        n.sw_dict(beta=1.0, err=True)

    with pytest.raises(ValueError):
        n.lam_dbeta(beta=1.0, err=True)

    with pytest.raises(ValueError):
        n.secondvirial_sw(beta=1.0, err=True)

    assert isinstance(n.sw_dict(1.0), dict)


def test_nf_sw() -> None:
    cols = ["sig", "eps", "lam", "B2"]
    table = pd.read_csv(data / "eff_sw_.csv").assign(beta=lambda x: 1.0 / x["temp"])

    if len(table) > nsamp:
        table = table.sample(nsamp)

    for (sig, eps, lam), g in table.groupby(["sig", "eps", "lam"]):
        p = pots.SquareWell(sig=sig, eps=eps, lam=lam)

        a = NoroFrenkelPair(phi=p.phi, segments=p.segments, r_min=sig, phi_min=eps)

        out = a.table(g["beta"], props=None)  # type: ignore[arg-type]

        for col in cols:
            left = col + "_eff"
            np.testing.assert_allclose(g[left], out[col])


def test_nf_lj() -> None:
    cols = ["sig", "eps", "lam", "B2", "sig_dbeta"]

    table = (
        pd.read_csv(data / "eff_lj_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
        .assign(sig_dbeta_eff=lambda x: -x["dsig_eff"])
    )

    for g, a in iter_phi_lj(table, nsamp):
        out = a.table(g["beta"], cols)  # type: ignore[arg-type]

        for col in cols:
            left = col + "_eff"
            np.testing.assert_allclose(g[left], out[col], rtol=1e-3)


def test_nf_nm() -> None:
    cols = ["sig", "eps", "lam", "B2"]
    table = (
        pd.read_csv(data / "eff_lj_nm_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
    )

    if len(table) > nsamp:
        table = table.sample(nsamp)
    sig = eps = 1.0

    for (n_exp, m_exp), g in table.groupby(["n_exp", "m_exp"]):
        p = pots.LennardJonesNM(n=n_exp, m=m_exp, sig=sig, eps=eps)

        a = NoroFrenkelPair.from_phi(
            phi=p.phi, segments=p.segments, r_min=sig, bounds=[0.5, 1.5]
        )

        out = a.table(g["beta"], cols)  # type: ignore[arg-type]

        for col in cols:
            left = col + "_eff"
            np.testing.assert_allclose(g[left], out[col], rtol=1e-3)


def test_nf_yk() -> None:
    cols = ["sig", "eps", "lam", "B2"]

    table = (
        pd.read_csv(data / "eff_yukawa_.csv")
        .assign(beta=lambda x: 1.0 / x["temp"])
        .drop_duplicates()
    )

    if len(table) > nsamp:
        table = table.sample(nsamp)
    sig = eps = 1.0

    for (z_yukawa), g in table.groupby("z_yukawa"):
        p = pots.Yukawa(z=z_yukawa, sig=sig, eps=eps)  # type: ignore[arg-type]

        a = NoroFrenkelPair(phi=p.phi, segments=p.segments, r_min=sig, phi_min=-1.0)

        out = a.table(g["beta"], cols)  # type: ignore[arg-type]

        for col in cols:
            left = col + "_eff"
            np.testing.assert_allclose(g[left], out[col], rtol=1e-3)



"""
Top level API :mod:`analphipy`
==============================

The top level API provides Pair potentials and analysis routines.
"""

from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _version

from . import measures, norofrenkel, potential

try:
    __version__ = _version("analphipy")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "999"

__author__ = """William P. Krekelberg"""
__email__ = "wpk@nist.gov"

__all__ = [
    # "PhiAnalytic",
    # "PhiGeneric",
    # "LennardJones",
    # "LennardJonesNM",
    # "SquareWell",
    # "HardSphere",
    # "Yukawa",
    # "CubicTable",
    # "factory_phi",
    # "NoroFrenkelPair",
    # "measures",
    # "norofrenkel",
    "__version__",
    "measures",
    "norofrenkel",
    # "PhiBaseCuttable",
    # "PhiBaseGenericCut",
    "potential",
]



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
from typing import TYPE_CHECKING

import numpy as np
import pytest

import analphipy.potential as pots
from analphipy.norofrenkel import NoroFrenkelPair

if TYPE_CHECKING:
    from analphipy.base_potential import PhiAbstract


def _do_test(nf, beta, prop, dprop, dx=1e-8, rtol=1e-3) -> None:
    a = getattr(nf, dprop)(beta)

    f = getattr(nf, prop)

    # central difference:
    b = (f(beta + dx * 0.5) - f(beta - dx * 0.5)) / dx

    np.testing.assert_allclose(a, b, rtol=rtol)


def factory_lj(lfs=False, cut=False, **kws):
    rcut = kws.pop("rcut") if lfs or cut else None

    p: PhiAbstract

    p = pots.LennardJones(**kws)

    if cut and lfs:
        msg = "Can only specify one of lfs and cut"
        raise ValueError(msg)

    if cut:
        assert isinstance(rcut, float)
        return p.cut(rcut=rcut)

    if lfs:
        assert isinstance(rcut, float)
        return p.lfs(rcut=rcut)

    return p


BETAS = [0.1, 0.5, 1.0]


def test_nf_deriv_lj() -> None:
    sig = eps = 1.0

    p = pots.LennardJones(sig=sig, eps=eps)

    nf = NoroFrenkelPair.from_phi(
        p.phi, p.segments, r_min=sig, bounds=[0.5 * sig, 1.5 * sig]
    )

    for prop in ("sig", "B2", "lam"):
        for beta in BETAS:
            _do_test(nf, beta, prop, prop + "_dbeta", dx=1e-8, rtol=1e-2)


@pytest.mark.parametrize("rcut", [2.0, 3.0, 5.0])
def test_nf_deriv_lj_cut(rcut) -> None:
    sig = eps = 1.0

    p = pots.LennardJones(sig=sig, eps=eps).cut(rcut)

    nf = NoroFrenkelPair.from_phi(
        p.phi, p.segments, r_min=sig, bounds=[0.5 * sig, 1.5 * sig]
    )

    for prop in ("sig", "B2", "lam"):
        for beta in BETAS:
            _do_test(nf, beta, prop, prop + "_dbeta", dx=1e-8, rtol=1e-2)


@pytest.mark.parametrize("z", [1.0, 2.0, 4.0])
def test_nf_deriv_yk(z) -> None:
    sig = eps = 1.0

    p = pots.Yukawa(sig=sig, eps=eps, z=z)

    nf = NoroFrenkelPair(p.phi, p.segments, r_min=sig, phi_min=-1.0)

    for prop in ("sig", "B2", "lam"):
        for beta in BETAS:
            _do_test(nf, beta, prop, prop + "_dbeta", dx=1e-8, rtol=1e-2)
